# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from contextlib import contextmanager
import copy
from textwrap import wrap

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

import bayeslite.core
from bayeslite.exception import BayesLiteException as BLE
import bdbcontrib.bql_utils as bqlu
from bdbcontrib.population_method import population_method

###############################################################################
###                                   PUBLIC                                ###
###############################################################################

@population_method(population_to_bdb=0, generator_name=1)
def mi_hist(bdb, generator_name, col1, col2, num_samples=1000, bins=5):
    """
    Histogram of estimated mutual information between the two columns for each
    of a generator's model instances.

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        Active BayesDB instance.
    generator_name : str
        Name of the generator to compute MI(col1;col2)
    col1, col2 : str
        Name of the columns to compute MI(col1;col2)
    num_samples : int, optional
        Number of samples to use in the Monte Carlo estimate of MI.
    bins : int, optional
        Number of bins in the histogram.

    Returns
    -------
    figure : matplotlib.figure.Figure
    """
    generator_id = bayeslite.core.bayesdb_get_generator(bdb, generator_name)
    bql = '''
        SELECT COUNT(modelno) FROM bayesdb_generator_model
            WHERE generator_id = ?
    '''
    counts = bdb.execute(bql, (generator_id,))
    num_models = counts.fetchvalue()
    if num_models == 0:
        raise ValueError('No models to plot mutual information over!')

    figure, ax = plt.subplots(figsize=(6, 6))

    mis = []
    for modelno in range(num_models):
        bql = '''
            ESTIMATE MUTUAL INFORMATION OF {} WITH {} USING {} SAMPLES FROM {}
                USING MODEL {} LIMIT 1
        '''.format(col1, col2, num_samples, generator_name, modelno)
        cursor = bdb.execute(bql)
        mutual_information = cursor.fetchvalue()
        mis.append(mutual_information)
    ax.hist(mis, bins, normed=True)
    ax.set_xlabel('Mutual Information')
    ax.set_ylabel('Density')
    ax.set_title('Mutual information of {} with {}'.format(col1, col2))

    return figure

@population_method(population_to_bdb=0, specifier_to_df=1,
                   generator_name='generator_name')
def pairplot(bdb, df, generator_name=None, stattypes=None, show_contour=False,
             colorby=None, show_missing=False, show_full=False,
             pad=None, h_pad=None, **kwargs):
    """Plot array of plots for all pairs of columns.

    Plots continuous-continuous pairs as scatter (optional KDE contour).
    Plots continuous-categorical pairs as violinplot.
    Plots categorical-categorical pairs as heatmap.

    Parameters
    ----------
    bdb : __population_to_bdb__
    df : __specifier_to_df__
    generator_name : __generator_name__ - used to find stattypes.
    stattypes : dict, optional {column_name: "categorical"|"numerical"}
        If you do not specify a generator name, have a column that the
        generator doesn't know about, or would like to override the
        statistical types the generator has for a given variable, then pass
        this dict of column names to types.
    show_contour : bool, optional
        If True, KDE contours are plotted on top of scatter plots
        and histograms.
    show_missing : bool, optional
        If True, rows with one missing value are plotted as lines on scatter
        plots.
    colorby : str, optional
        Name of a column to use to color data points in histograms and scatter
        plots.
    show_full : bool, optional
        Show full pairwise plots, rather than only lower triangular plots.
    pad : number, optional
        Adjust the vertical padding between plot components.
    h_pad : number, optional
        Adjust the horizontal padding between plot components.
    **kwargs : dict, optional
        Options to pass through to underlying plotter for pairs.

    Returns
    -------
    figure : matplotlib.figure.Figure
    """
    figure = _pairplot(df, bdb=bdb, generator_name=generator_name,
        stattypes=stattypes, show_contour=show_contour, colorby=colorby,
        show_missing=show_missing, show_full=show_full, **kwargs)
    if pad is not None or h_pad is not None:
      default_padding = -0.4 * (df.shape[1] - 2)
      if pad is None:
        pad = default_padding
      if h_pad is None:
        h_pad = default_padding
      figure.tight_layout(h_pad=h_pad, pad=pad)
    inches = len(df.columns) * 4
    figure.set_size_inches((inches, inches))

    return figure

@population_method(population_to_bdb=0, generator_name='generator_name',
                   population_name='population_name')
def pairplot_vars(bdb, varnames, colorby=None, generator_name=None,
                  population_name=None, **kwargs):
  """Use pairplot to show the given variables.

  See help(pairplot) for more plot options.

  Parameters
  ----------
  bdb: __population_to_bdb__
  varnames: list of one or more variables to plot.
  generator_name: __generator_name__
  population_name: __population_name__
  colorby: categorical variable to color all of the plots by.

  Returns
  -------
  figure: a matplotlib.figure.Figure
  """
  if len(varnames) < 1:
    raise BLE(ValueError('Pairplot at least one variable.'))
  qvars = varnames if colorby is None else set(varnames + [colorby])
  query_columns = '''"%s"''' % '''", "'''.join(qvars)
  bql = '''SELECT %s FROM %s''' % (query_columns, population_name)
  df = bqlu.query(bdb, bql)
  return pairplot(bdb, df, generator_name=generator_name,
      colorby=colorby, **kwargs)

@population_method(population_to_bdb=0, specifier_to_df=1)
def histogram(bdb, df, nbins=15, bins=None, normed=None):
    """Plot histogram of one- or two-column table.

    If two-column, subdivide the first column according to labels in
    the second column

    Parameters
    ----------
    bdb : __population_to_bdb__
    df : __specifier_to_df__
    nbins : int, optional
        Number of bins in the histogram.
    normed : bool, optional
        If True, normalizes the the area of the histogram (or each
        sub-histogram if df has two columns) to 1.

    Returns
    ----------
    figure: matplotlib.figure.Figure
    """
    df = df.dropna()
    if len(df.columns) == 0:
        raise BLE(ValueError('Tried to plot a histogram of an empty result.'))

    vartype = get_bayesdb_col_type(df.columns[0], df[df.columns[0]], bdb=bdb)
    if vartype == 'categorical':
        raise BLE(TypeError(
            "Cannot histogram categorical varible %s. Barplot? Colorby?" %
            (df.columns[0],)))
    if nbins is None:
        nbins = len(bins) if bins is not None else 15
    if bins is None:
        a = min(df.ix[:, 0].values)
        b = max(df.ix[:, 0].values)
        support = b - a
        interval = support / nbins
        bins = np.linspace(a, b+interval, nbins)

    colorby = None
    if len(df.columns) > 1:
        if len(df.columns) > 2:
            raise BLE(ValueError('Got more columns than data and colorby.'))
        colorby = df.columns[1]
        colorby_stattype = get_bayesdb_col_type(
            df.columns[1], df[df.columns[1]], bdb=bdb)
        if colorby_stattype != 'categorical':
            raise BLE(TypeError("Cannot color by non-categorical variable " +
                                colorby))
        colorby_vals = df[colorby].unique()

    figure, ax = plt.subplots(tight_layout=False, facecolor='white')
    if colorby is None:
        ax.hist(df.ix[:, 0].values, bins=bins, color='#383838',
            edgecolor='none', normed=normed)
        plot_title = df.columns[0]
    else:
        colors = sns.color_palette('deep', len(colorby_vals))
        for color, cbv in zip(colors, colorby_vals):
            subdf = df[df[colorby] == cbv]
            ax.hist(subdf.ix[:, 0].values, bins=bins, color=color, alpha=.5,
                edgecolor='none', normed=normed, label=str(cbv))
        ax.legend(loc=0, title=colorby)
        plot_title = df.columns[0] + " by " + colorby

    if normed:
        plot_title += " (normalized)"

    ax.set_title(plot_title)
    ax.set_xlabel(df.columns[0])
    return figure

@population_method(population_to_bdb=0, specifier_to_df=1)
def barplot(bdb, df):
    """Make bar-plot from categories and their heights.

    First column specifies names; second column specifies heights.

    Parameters
    ----------
    bdb : __population_to_bdb__
    df : __specifier_to_df__

    Returns
    ----------
    figure: matplotlib.figure.Figure
    """
    if df.shape[1] != 2:
        raise BLE(ValueError(
            'Need two columns of output from SELECT for barplot.'))
    height_inches = df.shape[0] / 2.0
    figure, ax = plt.subplots(figsize=(height_inches, 5))

    ax.bar([x - .5 for x in range(df.shape[0])], df.ix[:, 1].values,
        color='#333333', edgecolor='#333333')
    ax.set_xticks(range(df.shape[0]))
    ax.set_xticklabels(df.ix[:, 0].values, rotation=90)
    ax.set_xlim([-1, df.shape[0] - .5])
    ax.set_ylabel(df.columns[1])
    ax.set_xlabel(df.columns[0])

    return figure

###############################################################################
###                              INTERNAL                                   ###
###############################################################################

MODEL_TO_TYPE_LOOKUP = {
    'normal_inverse_gamma': 'numerical',
    'symmetric_dirichlet_discrete': 'categorical',
}


def rotate_tick_labels(ax, axis=None, rotation=None):
    """Rotate tick labels on one axis if specified, or both otherwise.

    axis : str 'x' or 'y' or None
        If None, then both axes are rotated as requested.
    rotation : number or 'horizontal', 'vertical', 'parallel', 'perpendicular'.
        Angle or orientation to rotate to.

    Returns: ax
    """
    if rotation is None:
        rotation = 'perpendicular'
    if axis:
        labels = ax.get_xticklabels() if axis == 'x' else ax.get_yticklabels()
        for tick in labels:
            if rotation == 'by_length':
                tick_rotation = 'horizontal'
                if axis == 'x' and len(tick.get_text()) > 4:
                    tick_rotation = 'vertical'
            elif rotation == 'parallel':
                tick_rotation = 'horizontal' if axis == 'x' else 'vertical'
            elif rotation == 'perpendicular':
                tick_rotation = 'vertical' if axis == 'x' else 'horizontal'
            else:
                tick_rotation = rotation
            tick.set_rotation(tick_rotation)
    else:
        rotate_tick_labels(ax, 'x', rotation)
        rotate_tick_labels(ax, 'y', rotation)
    return ax

def gen_collapsed_legend_from_dict(hl_colors_dict, loc=0, title=None,
        fontsize='medium', wrap_threshold=1000):
    """Creates a legend with entries grouped by color.

    For example, if a plot has multiple labels associated with the same color
    line, instead of generating a legend entry for each label, labels with the
    same colored line will be collapsed into longer, comma-separated labels.

    Parameters
    ----------
    hl_colors_dict : dict
        A dict of label, color pairs. Colors can be strings e.g. 'deeppink' or
        rgb or rgba tuples.
    loc : matplotlib compatible
        any matpltotlib-compbatible legend location identifier
    title : str
        legend title
    fontsize : int
        legend entry and title fontsize
    wrap_threshold : int
        max number of characters before wordwrap

    Returns
    -------
    legend : matplotlib.legend
    """
    if not isinstance(hl_colors_dict, dict):
        raise BLE(TypeError("hl_colors_dict must be a dict"))

    colors = list(set(hl_colors_dict.values()))
    from collections import defaultdict
    collapsed_dict = defaultdict(list)

    for label, color in hl_colors_dict.iteritems():
        collapsed_dict[color].append(str(label))

    for color in collapsed_dict.keys():
        collapsed_dict[color] = "\n".join(wrap(", ".join(
            sorted(collapsed_dict[color])), wrap_threshold))

    legend_artists = []
    legend_labels = []
    for color, label in collapsed_dict.iteritems():
        legend_artists.append(plt.Line2D((0, 1), (0, 0), color=color, lw=3))
        legend_labels.append(label)

    legend = plt.legend(legend_artists, legend_labels, loc=loc, title=title,
        fontsize=fontsize)

    return legend


def get_bayesdb_col_type(column_name, df_column, bdb=None,
                         generator_name=None):
    # If column_name is a column label (not a short name!) then the modeltype
    # of the column will be returned otherwise we guess.

    if isinstance(df_column, pd.DataFrame):
        raise BLE(TypeError(
            'Multiple columns in the query result have the same name (%s).' %
            (column_name,)))

    def guess_column_type(df_column):
        pd_type = df_column.dtype
        if pd_type is str or pd_type == np.object:
            return 'categorical'
        else:
            if len(df_column.unique()) < 30:
                return 'categorical'
            else:
                return 'numerical'

    if bdb is not None and generator_name is not None:
        try:
            coltype = bqlu.get_column_stattype(bdb, generator_name, column_name)
            # XXX: Force cyclic -> numeric because there is no need to plot
            # cyclic data any differently until we implement rose plots. See
            # http://matplotlib.org/examples/pie_and_polar_charts/polar_bar_demo.html
            # for an example.
            if coltype.lower() == 'cyclic':
                coltype = 'numerical'
            return coltype
        except IndexError:
            return guess_column_type(df_column)
    else:
        return guess_column_type(df_column)


def conv_categorical_vals_to_numeric(data_srs):
    # TODO: get real valuemap from btable
    unique_vals = sorted(data_srs.unique().tolist())
    lookup = dict(zip(unique_vals, range(len(unique_vals))))
    values = data_srs.values.tolist()
    for i, x in enumerate(values):
        values[i] = lookup[x]
    return np.array(values, dtype=float), unique_vals, lookup


# FIXME: STUB
def prep_plot_df(data_df, var_names):
    return data_df[list(var_names)]

def drop_inf_and_nan(np_Series):
    return np_Series.replace([-np.inf, np.inf], np.nan).dropna()

def seaborn_broken_bins(np_Series):
    # XXX: Seaborn is broken, uses np.asarray(values).squeeze()
    # which for lists of one is an "unsized object" from which you
    # cannot calculate freedman diaconis bins.
    if len(np_Series) == 1:
        return []
    # When seaborn uses its _freedman_diaconis_bins it forgets to np.ceil,
    # so do that for it. Again (this has been broken and fixed before). :-p
    # This is being inserted at seaborn==0.6.0.
    return np.ceil(sns.distributions._freedman_diaconis_bins(np_Series))

def do_hist(data_srs, **kwargs):
    ax = kwargs.get('ax', None)
    bdb = kwargs.get('bdb', None)
    dtype = kwargs.get('dtype', None)
    show_contour = kwargs.get('show_contour', None)
    generator_name = kwargs.get('generator_name', None)
    colors = kwargs.get('colors', None)

    if dtype is None:
        dtype = get_bayesdb_col_type(data_srs.columns[0], data_srs, bdb=bdb,
            generator_name=generator_name)

    if ax is None:
        ax = plt.gca()

    if len(data_srs.shape) > 1:
        if colors is None and data_srs.shape[1] != 1:
            raise BLE(ValueError(
                'If a dummy column is specified,'
                ' colors must also be specified.'))

    data_srs = data_srs.dropna()

    if dtype == 'categorical':
        vals, uvals, _ = conv_categorical_vals_to_numeric(data_srs.ix[:, 0])
        if colors is not None:
            color_lst = []
            stacks = []
            for val, color in colors.iteritems():
                subval = vals[data_srs.ix[:, 1].values == val]
                color_lst.append(color)
                stacks.append(subval)
            ax.hist(stacks, bins=len(uvals), color=color_lst, alpha=.9,
                histtype='barstacked', rwidth=1.0)
        else:
            ax.hist(vals, bins=len(uvals))
        ax.set_xticks(range(len(uvals)))
        ax.set_xticklabels(uvals)
    else:
        do_kde = show_contour
        if colors is not None:
            for val, color in colors.iteritems():
                subdf = data_srs.loc[data_srs.ix[:, 1] == val]
                values = drop_inf_and_nan(subdf.ix[:, 0])
                if len(values) < 2: # Then seaborn would break. :-p
                    continue
                bins = seaborn_broken_bins(values)
                sns.distplot(values, kde=do_kde, ax=ax, color=color, bins=bins)
        else:
            values = drop_inf_and_nan(data_srs)
            bins = seaborn_broken_bins(values)
            sns.distplot(values, kde=do_kde, ax=ax, bins=bins)

    return ax


def do_heatmap(plot_df, unused_vartypes, **kwargs):
    ax = kwargs.get('ax', None)

    plot_df = plot_df.dropna()
    if ax is None:
        ax = plt.gca()

    vals_x, uvals_x, _ = conv_categorical_vals_to_numeric(plot_df.ix[:, 0])
    vals_y, uvals_y, _ = conv_categorical_vals_to_numeric(plot_df.ix[:, 1])

    bins_x = len(uvals_x)
    bins_y = len(uvals_y)

    hst, _, _ = np.histogram2d(vals_y, vals_x, bins=[bins_y, bins_x])
    cmap = 'BuGn'
    ax.matshow(hst, aspect='auto', origin='lower', cmap=cmap)
    ax.xaxis.set_tick_params(labeltop='off', labelbottom='on') # RESETS LABELS!
    ax.grid(b=False)
    ax.set_xticks(range(bins_x))
    ax.set_yticks(range(bins_y))
    ax.set_xticklabels(uvals_x)
    ax.set_yticklabels(uvals_y)
    rotate_tick_labels(ax)
    return ax


def do_violinplot(plot_df, vartypes, **kwargs):
    ax = kwargs.get('ax', None)
    colors = kwargs.get('colors', None)
    cut=1
    if kwargs.get('show_contour', None) == False:
        cut=0
    for handled in ('ax', 'show_contour', 'colors'):
        if handled in kwargs:
            del kwargs[handled]

    dummy = plot_df.shape[1] == 3

    plot_df = plot_df.dropna()
    if ax is None:
        ax = plt.gca()
    assert vartypes[0] != vartypes[1]
    vert = vartypes[1] == 'numerical'
    plot_df = copy.deepcopy(plot_df.dropna())
    if vert:
        groupby = plot_df.columns[0]
    else:
        groupby = plot_df.columns[1]

    _, unique_vals, _ = conv_categorical_vals_to_numeric(plot_df[groupby])

    unique_vals = np.sort(unique_vals)
    n_vals = len(plot_df[groupby].unique())
    if dummy:
        sub_vals = np.sort(plot_df[groupby].unique())
        axis = sns.violinplot(
            x=plot_df.columns[0],
            y=plot_df.columns[1],
            data=plot_df,
            order=sub_vals, hue=plot_df.columns[2],
            names=sub_vals, ax=ax, orient=("v" if vert else "h"),
            palette=colors, inner='quartile', cut=cut, **kwargs)
        axis.legend_ = None
    else:
        sns.violinplot(
            x=plot_df.columns[0],
            y=plot_df.columns[1],
            data=plot_df,
            order=unique_vals, names=unique_vals, ax=ax,
            orient=("v" if vert else "h"),
            color='SteelBlue', inner='quartile', cut=cut, **kwargs)

    if vert:
        ax.set_xlim([-.5, n_vals-.5])
        ax.set_xticks(np.arange(n_vals))
        ax.set_xticklabels(unique_vals)
    else:
        ax.set_ylim([-.5, n_vals-.5])
        ax.set_yticks(np.arange(n_vals))
        ax.set_yticklabels(unique_vals)

    return ax


def do_kdeplot(plot_df, unused_vartypes, bdb=None, generator_name=None,
               **kwargs):
    # XXX: kdeplot is not a good choice for small amounts of data because
    # it uses a kernel density estimator to crease a smooth heatmap. On the
    # other hand, scatter plots are uninformative given lots of data---the
    # points get jumbled up. We may just want to set a threshold (N=100)?

    _unused_generator_name = generator_name  # So it doesn't end up in kwargs
    _unused_bdb = bdb

    ax = kwargs.get('ax', None)
    show_contour = kwargs.get('show_contour', False)
    colors = kwargs.get('colors', None)
    show_missing = kwargs.get('show_missing', False)
    for handled in ('ax', 'colors', 'show_contour', 'show_missing'):
        if handled in kwargs:
            del kwargs[handled]

    if ax is None:
        ax = plt.gca()

    xlim = [plot_df.ix[:, 0].min(), plot_df.ix[:, 0].max()]
    ylim = [plot_df.ix[:, 1].min(), plot_df.ix[:, 1].max()]

    dummy = plot_df.shape[1] == 3

    null_rows = plot_df[plot_df.isnull().any(axis=1)]
    df = plot_df.dropna()

    if not dummy:
        plt.scatter(df.values[:, 0], df.values[:, 1], alpha=.5,
            color='steelblue', zorder=2, **kwargs)
        # plot nulls
        if show_missing:
            nacol_x = null_rows.ix[:, 0].dropna()
            for x in nacol_x.values:
                plt.plot([x, x], ylim, color='crimson', alpha=.2, lw=1,
                    zorder=1)
            nacol_y = null_rows.ix[:, 1].dropna()
            for y in nacol_y.values:
                plt.plot(xlim, [y, y], color='crimson', alpha=.2, lw=1,
                    zorder=1)
    else:
        assert isinstance(colors, dict)
        for val, color in colors.iteritems():
            subdf = df.loc[df.ix[:, 2] == val]
            if subdf.shape[0] == 0:
                continue
            plt.scatter(subdf.values[:, 0], subdf.values[:, 1], alpha=.5,
                color=color, zorder=2, **kwargs)
            subnull = null_rows.loc[null_rows.ix[:, 2] == val]
            if show_missing:
                nacol_x = subnull.ix[:, 0].dropna()
                for x in nacol_x.values:
                    plt.plot([x, x], ylim, color=color, alpha=.3, lw=2,
                        zorder=1)
                nacol_y = subnull.ix[:, 1].dropna()
                for y in nacol_y.values:
                    plt.plot(xlim, [y, y], color=color, alpha=.3, lw=2,
                        zorder=1)

    if show_contour:
        try:
            sns.kdeplot(df.ix[:, :2].values, ax=ax, **kwargs)
        except ValueError:
            # Displaying a plot without a requested contour is better
            # than crashing.
            pass

            # This actually happens: with the 'skewed_numeric_5'
            # distribution in tests/test_plot_utils.py, we get:
            # seaborn/distributions.py:597: in kdeplot
            #     ax, **kwargs)
            # seaborn/distributions.py:380: in _bivariate_kdeplot
            #     cset = contour_func(xx, yy, z, n_levels, **kwargs)
            # matplotlib/axes/_axes.py:5333: in contour
            #     return mcontour.QuadContourSet(self, *args, **kwargs)
            # matplotlib/contour.py:1429: in __init__
            #     ContourSet.__init__(self, ax, *args, **kwargs)
            # matplotlib/contour.py:876: in __init__
            #     self._process_levels()
            # matplotlib/contour.py:1207: in _process_levels
            #     self.vmin = np.amin(self.levels)
            # numpy/core/fromnumeric.py:2224: in amin
            #     out=out, keepdims=keepdims)
            # _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _
            # a = array([], dtype=float64),
            # axis = None, out = None, keepdims = False
            #
            #     def _amin(a, axis=None, out=None, keepdims=False):
            # >       return umr_minimum(a, axis, None, out, keepdims)
            # E       ValueError: zero-size array to reduction operation
            #                     minimum which has no identity

    return ax


# No support for cyclic at this time
DO_PLOT_FUNC = dict()
DO_PLOT_FUNC[('categorical', 'categorical')] = do_heatmap
DO_PLOT_FUNC[('categorical', 'numerical')] = do_violinplot
DO_PLOT_FUNC[('numerical', 'categorical')] = do_violinplot
DO_PLOT_FUNC[('numerical', 'numerical')] = do_kdeplot


def do_pair_plot(plot_df, vartypes, **kwargs):
    # determine plot_types
    if kwargs.get('ax', None) is None:
        kwargs['ax'] = plt.gca()

    ax = DO_PLOT_FUNC[vartypes](plot_df, vartypes, **kwargs)
    return ax

def ensure_full_square(data_df, pivot_kws, default_value=0):
    """Ensure that all pairs are bidirectionally present in a pivot df.

    data_df is expected to have an index column, a columns column, and a
    data column, as specified by the pivot_kws (as they would be supplied
    to seaborn.clustermap).

    Creates a complete matrix where every value in index has an entry for
    every value in columns, and vice versa, filling with zeros as needed.

    Scipy has been known to segfault without this, or give a ValueError,
    depending on the version.

    Return: a new complete-square df with those three columns.
    """
    iname = pivot_kws['index']
    cname = pivot_kws['columns']
    vname = pivot_kws['values']

    import sets
    index_values = set(data_df[iname])
    columns_values = set(data_df[cname])
    if len(index_values) == 0 or len(columns_values) == 0:
        return pd.DataFrame()
    data = {}
    for index, row in data_df.iterrows():
        if row[iname] not in data:
            data[row[iname]] = {}
        sub = data[row[iname]]
        if row[cname] not in sub:
            sub[row[cname]] = row[vname]
    rows = []
    for ind in sorted(index_values):
        for col in sorted(columns_values):
            if ind in data and col in data[ind]:
                rows.append([ind, col, data[ind][col]])
            else:
                rows.append([ind, col, default_value])
    new_df = pd.DataFrame(rows)
    new_df.columns = [iname, cname, vname]
    return new_df

@population_method(specifier_to_df=0)
def heatmap(data_df, row_ordering=None, col_ordering=None, **kwargs):
    """Plot heatmaps, optionally clustered.

    Parameters
    ----------
    deps : __specifier_to_df__
        Must have two categorical columns and a numeric column.
        The format is assumed to be that the numeric column of values is last,
        and the two categorical columns are immediately before that.

        The canonical example of that kind of data is the result of an
        ESTIMATE ... PAIRWISE query, estimating dependence probability,
        mutual information, correlation, etc.

        If your columns are not at the end that way, then pass pivot_kws too.

    row_ordering, col_ordering : list<int>
        Specify the order of labels on the x and y axis of the heatmap.
        If these are specified, we will not call try to cluster a different way,
        and return a plain heatmap.

        To access the row and column indices from a clustermap object, use:
        clustermap.dendrogram_row.reordered_ind (for rows)
        clustermap.dendrogram_col.reordered_ind (for cols)

    **kwargs :
        to pass to seaborn.clustermap. See seaborn documentation. Of particular
        importance is the `pivot_kws` kwarg. `pivot_kws` is a dict with entries
        index, column, and values that let clustermap know how to reshape the
        data. If the query does not follow the standard ESTIMATE PAIRWISE
        output, it may be necessary to define `pivot_kws`.
        Other keywords here include vmin and vmax, linewidths, figsize, etc.

    Returns
    -------
    clustermap: seaborn.clustermap
    """
    clustermap_kws = {}
    clustermap_kws.update(kwargs)
    data_df.fillna(0, inplace=True)
    if clustermap_kws is None:
        clustermap_kws = {}
    if 'figsize' not in clustermap_kws:
        half_root_col = (data_df.shape[0] ** .5) / 2.0
        clustermap_kws['figsize'] = (half_root_col, .8 * half_root_col)
    if 'linewidths' not in clustermap_kws:
        clustermap_kws['linewidths'] = 0.2
    if 'pivot_kws' not in clustermap_kws:
        # XXX: If the user doesn't tell us otherwise, we assume that this comes
        # from a standard estimate pairwise query, which outputs columns
        # (table_id, col0, col1, value). The indices are indexed from the back
        # because it will also handle the no-table_id case
        data_df.columns = [' '*i for i in range(1, len(data_df.columns))] + \
            ['value']
        pivot_kws = {
            'index': data_df.columns[-3],
            'columns': data_df.columns[-2],
            'values': data_df.columns[-1],
        }
        clustermap_kws['pivot_kws'] = pivot_kws

    if clustermap_kws.get('cmap', None) is None:
        # Choose a soothing blue colormap
        clustermap_kws['cmap'] = 'BuGn'

    autoset_vlimits = False
    if 'vmin' not in clustermap_kws:
        autoset_vlimits = True
        clustermap_kws['vmin'] = \
            data_df[clustermap_kws['pivot_kws']['values']].min()
    if 'vmax' not in clustermap_kws:
        autoset_vlimits = True
        clustermap_kws['vmax'] = \
            data_df[clustermap_kws['pivot_kws']['values']].max()
    if (autoset_vlimits and
        clustermap_kws['vmin'] >= 0 and
        clustermap_kws['vmax'] <= 1):
        clustermap_kws['vmin'] = 0
        clustermap_kws['vmax'] = 1
    if autoset_vlimits:
        print ("Detected value limits as [%s, %s]. Override with vmin and vmax."
               % (clustermap_kws['vmin'], clustermap_kws['vmax']))


    if row_ordering is not None and col_ordering is not None:
        index = clustermap_kws['pivot_kws']['index']
        columns = clustermap_kws['pivot_kws']['columns']
        values = clustermap_kws['pivot_kws']['values']
        df = data_df.pivot(index, columns, values)
        df = df.ix[:, col_ordering]
        df = df.ix[row_ordering, :]
        del clustermap_kws['pivot_kws']
        del clustermap_kws['figsize']
        _fig, ax = plt.subplots()
        sns.heatmap(df, ax=ax, **clustermap_kws)
        rotate_tick_labels(ax)
        return (ax, row_ordering, col_ordering)
    else:
        complete_df = ensure_full_square(data_df, clustermap_kws['pivot_kws'])
        hmap = sns.clustermap(complete_df, **clustermap_kws)
        rotate_tick_labels(hmap.ax_heatmap)
        return hmap


# TODO: bdb, and table_name should be optional arguments
def _pairplot(df, bdb=None, generator_name=None, stattypes=None,
        show_contour=False, colorby=None, show_missing=False, show_full=False,
        **kwargs):
    """Plots the columns in data_df in a facet grid.

    Supports the following pairs:
    - categorical-categorical pairs are displayed as a heatmap
    - continuous-continuous pairs are displayed as a kdeplot
    - categorical-continuous pairs are displayed on a violin plot

    Parameters
    ----------
    df : pandas.DataFrame
        The input data---the result of a BQL/SQL query
    bdb : bayeslite.BayesDB (optional)
        The BayesDB object associated with `df`. Having the BayesDB object and
        the generator for the data allows pairplot to choose plot types.
    generator_name : str
        The name of generator associated with `df` and `bdb`.
    stattypes : dict, optional {column_name: "categorical"|"numerical"}
        If you do not specify a generator name, have a column that the
        generator doesn't know about, or would like to override the
        statistical types the generator has for a given variable, then pass
        this dict of column names to types.
    show_contour : bool
        If True, KDE contours are plotted on top of scatter plots
        and histograms.
    show_missing : bool
        If True, rows with one missing value are plotted as lines on scatter
        plots.
    colorby : str
        Name of a column to use to color data points in histograms and scatter
        plots.
    show_full : bool
        Show full pairwise plots, rather than only lower triangular plots.
    kwargs : dict
        Options to pass through to underlying plotting function (for pairs).

    Returns
    -------
    figure : matplotlib.figure.Figure
        A num_columns by num_columns Gridspec of pairplot axes.

    Notes
    -----
    Support soon for ordered continuous combinations. It may be best
    to plot all ordered continuous pairs as heatmap.
    """
    # NOTE:Things to consider:
    # - String values are a possibility (categorical)
    # - who knows what the columns are named. What is the user selects columns
    #   as shortname?
    # - where to handle dropping NaNs? Missing values may be informative.

    data_df = df
    if stattypes is None:
        stattypes = {}

    colors = None
    if colorby is not None:
        n_colorby = 0
        for colname in data_df.columns:
            # XXX: This is not guaranteed to work on all Unicode characters.
            if colorby.lower() == colname.lower():
                n_colorby += 1
                colorby = colname
        if n_colorby == 0:
            raise BLE(ValueError(
                'colorby column, {}, not found.'.format(colorby)))
        elif n_colorby > 1:
            raise BLE(ValueError('Multiple columns named, {}.'.format(colorby)))

        dummy = data_df[colorby].dropna()
        dvals = np.sort(dummy.unique())
        ndvals = len(dvals)
        dval_type = "categorical"
        if colorby in stattypes:
            dval_type = stattypes[colorby]
        elif generator_name:
            dval_type = get_bayesdb_col_type(colorby, dummy, bdb=bdb,
                                             generator_name=generator_name)
        if dval_type.lower() != 'categorical':
            raise BLE(ValueError('colorby columns must be categorical.'))
        cmap = sns.color_palette("Set1", ndvals)
        colors = {}
        for val, color in zip(dvals, cmap):
            colors[val] = color

    all_varnames = [c for c in data_df.columns if c != colorby]
    n_vars = len(all_varnames)
    plt_grid = gridspec.GridSpec(n_vars, n_vars)
    figure = plt.figure()

    # if there is only one variable, just do a hist
    if n_vars == 1:
        ax = plt.gca()
        varname = data_df.columns[0]
        vartype = "categorical"
        if varname in stattypes:
            vartype = stattypes[varname]
        elif generator_name:
            vartype = get_bayesdb_col_type(varname, data_df[varname], bdb=bdb,
                                           generator_name=generator_name)
        do_hist(data_df, dtype=vartype, ax=ax, bdb=bdb, colors=colors)
        rotate_tick_labels(ax)
        return

    xmins = np.ones((n_vars, n_vars))*float('Inf')
    xmaxs = np.ones((n_vars, n_vars))*float('-Inf')
    ymins = np.ones((n_vars, n_vars))*float('Inf')
    ymaxs = np.ones((n_vars, n_vars))*float('-Inf')

    vartypes = []
    for varname in all_varnames:
        vartype = "categorical"
        if varname in stattypes:
            vartype = stattypes[varname]
        elif generator_name:
            vartype = get_bayesdb_col_type(varname, data_df[varname], bdb=bdb,
                                           generator_name=generator_name)
        vartypes.append(vartype)

    # store each axes; reaccessing ax with plt.subplot(plt_grid[a,b]) may
    # overwrite the ax
    axes = [[] for _ in xrange(len(all_varnames))]
    for x_pos, var_name_x in enumerate(all_varnames):
        var_x_type = vartypes[x_pos]
        for y_pos, var_name_y in enumerate(all_varnames):
            var_y_type = vartypes[y_pos]

            ax = figure.add_subplot(plt_grid[y_pos, x_pos])
            axes[y_pos].append(ax)

            if x_pos == y_pos:
                varnames = [var_name_x]
                if colorby is not None:
                    varnames.append(colorby)
                ax = do_hist(data_df[varnames], dtype=var_x_type, ax=ax,
                    bdb=bdb, colors=colors, show_contour=show_contour)
            else:
                varnames = [var_name_x, var_name_y]
                vartypes_pair = (var_x_type, var_y_type,)
                if colorby is not None:
                    varnames.append(colorby)
                plot_df = prep_plot_df(data_df, varnames)
                ax = do_pair_plot(plot_df, vartypes_pair, ax=ax, bdb=bdb,
                    show_contour=show_contour, show_missing=show_missing,
                    colors=colors, **kwargs)

                ymins[y_pos, x_pos] = ax.get_ylim()[0]
                ymaxs[y_pos, x_pos] = ax.get_ylim()[1]
                xmins[y_pos, x_pos] = ax.get_xlim()[0]
                xmaxs[y_pos, x_pos] = ax.get_xlim()[1]

            ax.set_xlabel(var_name_x, fontweight='bold')
            ax.set_ylabel(var_name_y, fontweight='bold')

    for x_pos in range(n_vars):
        for y_pos in range(n_vars):
            ax = axes[y_pos][x_pos]

            # Self-histogram for x only, or comparative x against y:
            ax.set_xlim([np.min(xmins[:, x_pos]), np.max(xmaxs[:, x_pos])])
            if x_pos != y_pos:
                ax.set_ylim([np.min(ymins[y_pos, :]), np.max(ymaxs[y_pos, :])])

            # Y labels
            if x_pos > 0:
                if x_pos == n_vars - 1:  # All the way to the right:
                    ax.yaxis.tick_right()
                    ax.yaxis.set_label_position('right')
                else:  # No labels inside:
                    ax.set_ylabel('')
                    ax.set_yticklabels([])
            else:
                ax.yaxis.tick_left()
                ax.yaxis.set_label_position('left')

            # X labels:
            if y_pos < n_vars - 1:
                if y_pos == 0:  # At top, show x labels on top:
                    ax.xaxis.tick_top()
                    ax.xaxis.set_label_position('top')
                else:  # No labels inside:
                    ax.set_xlabel('')
                    ax.set_xticklabels([])
            else:
                ax.xaxis.tick_bottom()
                ax.xaxis.set_label_position('bottom')

            rotate_tick_labels(ax)

    def fake_axis_ticks(ax_tl, ax_tn):
        atl, btl = ax_tl.get_ylim()
        atn, btn = ax_tn.get_ylim()
        tnticks = ax_tn.get_yticks()
        yrange_tn = (btn-atn)
        yrange_tl = (btl-atl)
        tntick_ratios = [(t-atn)/yrange_tn for t in tnticks]
        ax_tl.set_yticks([r*yrange_tl+atl for r in tntick_ratios])
        ax_tl.set_yticklabels(tnticks)

    # Fix the top-left histogram y-axis ticks and labels.
    if show_full:
        fake_axis_ticks(axes[0][0], axes[0][1])
    fake_axis_ticks(axes[-1][-1], axes[-1][0])

    if colorby is not None:
        legend = gen_collapsed_legend_from_dict(colors, title=colorby)
        legend.draggable()

    # tril by default by deleting upper diagonal axes.
    if not show_full:
        for y_pos in range(n_vars):
            for x_pos in range(y_pos+1, n_vars):
                figure.delaxes(axes[y_pos][x_pos])

    return figure


def comparative_hist(df, bdb=None, nbins=15, normed=False):
    """Plot a histogram.

    Given a one-column pandas.DataFrame, df, plots a simple histogram. Given a
    two-column df plots the data in column one colored by an optional column 2.
    If given, column 2 must be categorical.

    Parameters
    ----------
    nbins : int
        Number of bins (bars)
    normed : bool
        If True, normalizes the the area of the histogram (or each
        sub-histogram if df has two columns) to 1.

    Returns
    -------
    figure: matplotlib.figure.Figure
    """
    df = df.dropna()

    vartype = get_bayesdb_col_type(df.columns[0], df[df.columns[0]], bdb=bdb)
    if vartype == 'categorical':
        values, labels, lookup = conv_categorical_vals_to_numeric(
            df[df.columns[0]])
        df.ix[:, 0] = values
        bins = len(labels)
        ticklabels = [0]*len(labels)
        for key, val in lookup.iteritems():
            ticklabels[val] = key
    else:
        a = min(df.ix[:, 0].values)
        b = max(df.ix[:, 0].values)
        support = b - a
        interval = support/nbins
        bins = np.linspace(a, b+interval, nbins)

    colorby = None
    if len(df.columns) > 1:
        if len(df.columns) > 2:
            raise BLE(NotImplementedError(
                'comparative_hist not defined on more than two variables.'))
        colorby = df.columns[1]
        colorby_vals = df[colorby].unique()

    figure, ax = plt.subplots(tight_layout=False, facecolor='white')
    if colorby is None:
        ax.hist(df.ix[:, 0].values, bins=bins, color='#383838',
            edgecolor='none', normed=normed)
        plot_title = df.columns[0]
    else:
        colors = sns.color_palette('deep', len(colorby_vals))
        for color, cbv in zip(colors, colorby_vals):
            subdf = df[df[colorby] == cbv]
            ax.hist(subdf.ix[:, 0].values, bins=bins, color=color, alpha=.5,
                edgecolor='none', normed=normed,
                label=("%s (n=%d)" % (str(cbv), len(subdf))))
        ax.legend(loc=0, title=colorby)
        plot_title = df.columns[0] + " by " + colorby

    if normed:
        plot_title += " (normalized)"

    ax.set_title(plot_title)
    ax.set_xlabel(df.columns[0])
    return figure



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import cgi
import sys
import traceback
import IPython.display
import IPython.utils.warn

from bayeslite.exception import BayesLiteException
from bayeslite.loggers import BqlLogger

class IpyLogger(BqlLogger):
  def info(self, msg_format, *values):
    message = msg_format % values
    html = ('<dib class="inner_cell" style="background-color:#F7F8E0">' +
            cgi.escape(message) + '</div>')
    IPython.display.display_html(html, raw=True)
  def warn(self, msg_format, *values):
    IPython.utils.warn.warn(msg_format % values)
  def exception(self, msg_format, *values):
    exc_type, exc_value, exc_traceback = sys.exc_info()
    if exc_type:
      if isinstance(exc_value, (BayesLiteException, ValueError)):
        IPython.utils.warn.error(msg_format + '\n' + str(exc_value), *values)
      else:
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        IPython.utils.warn.error('\n'.join(lines))
    else:
      self.warn("ERROR: " + msg_format, *values)

IPYTHON_LOGGER = IpyLogger()
def bayeslite_aware_render_traceback(self):
  IPYTHON_LOGGER.exception("")
  return []
BayesLiteException._render_traceback_ = bayeslite_aware_render_traceback



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# pylint: disable=no-member

import os
import pandas as pd
import re
import sys

import bayeslite
import bayeslite.core
from bayeslite.loggers import BqlLogger, logged_query
from bayeslite.exception import BayesLiteException as BLE

import bdbcontrib
from py_utils import helpsub

OPTFILE = "bayesdb-session-capture-opt.txt"

class Population(object):
  """Generative Population Model, wraps a BayesDB, and tracks one population.

  See Population.help() for a short menu of available methods.
  See help(Population) as usual for more complete information.
  """

  shortdocs = []

  @classmethod
  def method_imports(cls):
    """Runs decorators that add methods to Population."""
    # Set up a place for methods to deposit their short documentations for help:
    cls.shortdocs = []

    # These are here rather than in, say __init__.py so doing import bdbcontrib
    # just to get its __version__ for example doesn't need to run all that code.
    # __init__.py does import population (this file) for Population's __doc__
    # (same as this class's __init__.__doc__) so these can't just be at top.
    # But once you're using Populations, you have to pay this price once.
    import bql_utils
    import plot_utils
    import recipes
    import diagnostic_utils
    import crosscat_utils
    # Convenience alias:
    cls.q = cls.query
    cls.vartype = cls.get_column_stattype
    cls.quick_describe_columns = cls.variable_stattypes

  def __init__(self, name, csv_path=None, bdb_path=None, df=None, logger=None,
               session_capture_name=None):
    """Create a Population object, wrapping a bayeslite.BayesDB.

    name : str  REQUIRED.
        A name for the population, should use letters and underscores only.
        This will also be used as a table name in the bdb, and %t in queries
        will expand to this name. %g in queries will expand to the current
        population metamodel, also based on this name.
    csv_path : str
        The path to a comma-separated values file. If specified, will be used
        to populate the bdb. It must exist and be both readable and non-empty.
    df : pandas.DataFrame
        If specified, these data will be used to populate the bdb, superseding
        any csv_path. It must not be empty.
    bdb_path : str
        If specified, store data and analysis results here. If no other data
        source (csv or df) is specified, then it must already have been
        populated. If not specified, we will use a volatile in-memory bdb.
    logger : object
        Something on which we can call .info or .warn to send messages to the
        user. By default a bayeslite.loggers.BqlLogger, but could be QuietLogger
        (only results), SilentLogger (nothing), IpyLogger, CaptureLogger,
        LoggingLogger, or anything else that implements the BqlLogger interface.
    session_capture_name : String
        Signing up with your name and email and sending your session details
        to the MIT Probabilistic Computing Group helps build a community of
        support and helps improve your user experience. You can save your choice
        in a file called 'bayesdb-session-capture-opt.txt' in the directory
        where you run the software, or any parent directory. This option
        overrides any setting in such a file. Any string is interpreted as
        opting in to sending session details. False is interpreted as opting
        out. You must choose. If you choose to use an organization name or
        email, then please send a note to bayesdb@mit.edu to help us connect
        your sessions to you.

        If you encounter a bug, or something surprising, please include your
        session capture name in your report.

        If you opt out, you still allow us to count how often users opt out.

        DO NOT USE THIS SOFTWARE FOR HIPAA-COVERED, PERSONALLY IDENTIFIABLE,
        OR SIMILARLY SENSITIVE DATA! Opting out does not guarantee security.
    """
    Population.method_imports()
    assert re.match(r'\w+', name)
    assert df is not None or csv_path or bdb_path
    self.name = name
    self.generator_name = name + '_cc' # Because we use the default metamodel.
    self.csv_path = csv_path
    self.df = df
    self.bdb_path = bdb_path
    if logger is None:
      if 'IPython' in sys.modules:
        from bdbcontrib.loggers import IPYTHON_LOGGER as ipy
        self.logger = ipy
      else:
        self.logger = BqlLogger()
    else:
      self.logger = logger
    self.bdb = None
    self.status = None
    self.session_capture_name = None
    self.generators = []
    with logged_query('count-beacon', None, name='count-beacon'):
      self.initialize_session_capture(session_capture_name)
    self.initialize()

  def initialize_session_capture(self, name):
    if self.session_capture_name is not None:
      return
    if name is not None:
      self.session_capture_name = name
      return
    # Search for a session-capture name or opt-out saved as a file:
    searchdir = os.getcwd()
    while searchdir != os.path.dirname(searchdir):  # While not at root.
      try:
        with open(os.path.join(searchdir, OPTFILE), 'r') as optinfile:
          self.session_capture_name = optinfile.read()
          if self.session_capture_name == 'False':
            self.session_capture_name = False
          break
      except IOError:
        pass
      searchdir = os.path.dirname(searchdir)
    # No init option specified, no choice file found. Force the choice.
    if self.session_capture_name is None:
      raise BLE(
        "Please set session_capture_name option to Population.__init__\n"
        "  to either opt-in or opt-out of sending details of your usage of\n"
        "  this software to the MIT Probabilistic Computing Group.\n\n"
        "If you see this in one of our example notebooks,\n"
        "  return to the starting page, the Index.ipynb, to\n"
        "  make that choice.")

  def initialize(self):
    if self.bdb:
      self.check_representation()
      return
    self.bdb = bayeslite.bayesdb_open(self.bdb_path)
    if not bayeslite.core.bayesdb_has_table(self.bdb, self.name):
      if self.df is not None:
        bayeslite.read_pandas.bayesdb_read_pandas_df(
          self.bdb, self.name, self.df, create=True, ifnotexists=True)
      elif self.csv_path:
        bayeslite.bayesdb_read_csv_file(
          self.bdb, self.name, self.csv_path,
          header=True, create=True, ifnotexists=True)
      else:
        tables = self.list_tables()
        metamodels = self.list_metamodels()
        if len(tables) + len(metamodels) == 0:
          raise BLE(ValueError("No data sources specified, and an empty bdb."))
        else:
          raise BLE(ValueError("The name of the population must be the same"
                               " as a table in the bdb, one of: " +
                               ", ".join(tables) +
                               "\nNote also that the bdb has the following"
                               " metamodels defined: " + ", ".join(metamodels)))
    self.generators = self.query('''SELECT * FROM bayesdb_generator''')
    if len(self.generators) == 0:
      size = self.query('''SELECT COUNT(*) FROM %t''').ix[0, 0]
      assert 0 < size
      self.query('''
        CREATE GENERATOR %g IF NOT EXISTS FOR %t USING crosscat( GUESS(*) )''')
    self.check_representation()

  def check_representation(self):
    assert self.bdb, "Did you initialize?"
    assert self.session_capture_name is not None

  def interpret_bql(self, query_string):
    '''Replace %t and %g as appropriate.'''
    return re.sub(r'(^|(?<=\s))%t\b',
                  bayeslite.bql_quote_name(self.name),
                  re.sub(r'(^|(?<=\s))%g\b',
                         bayeslite.bql_quote_name(self.generator_name),
                         query_string))

  def sql_tracing(self, turn_on=True):
    """Trace underlying SQL, for debugging."""
    self.check_representation()
    # Always turn off:
    self.bdb.sql_untrace(self.bdb.sql_tracer)
    if turn_on:
      printer = lambda query, bindings: self.logger.info(
          "query: [%s] bindings: [%s]\n\n", query, bindings)
      self.bdb.sql_trace(printer)

  def reset(self):
    self.check_representation()
    self.query('drop generator if exists %s' % self.generator_name)
    self.query('drop table if exists %s' % self.name)
    self.bdb = None
    self.initialize()

  def specifier_to_df(self, spec):
    if isinstance(spec, pd.DataFrame):
      return spec
    else:
      return self.query(spec)

  def help(self, filter=None):
    """Show a short menu of available methods.

    filter : string or re
        Show only methods whose descriptions match the given pattern.
    """
    response = self.shortdocs
    if filter is not None:
      if hasattr(filter, '__call__'):
        response = response.filter(filter)
      else:
        response = [r for r in response if re.search(filter, r)]
    print '\n'.join(response)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# pylint: disable=no-member

# These methods are for analyses that involve some amount of
# data-analytical judgment encoded in the procedure and defaults, so
# they are not intended as primitives, but as suggested
# starting-points for a user's more meaningful analysis.

import pandas
import re

import bayeslite
from bayeslite.exception import BayesLiteException as BLE
import bdbcontrib
from bdbcontrib.population import Population
from bdbcontrib.population_method import population_method
from py_utils import helpsub

@population_method(population=0, generator_name='generator_name')
def quick_describe_variables(self, generator_name=None):
  """Show a univariate description of each variable."""
  assert generator_name
  if self.df is None or self.df.empty:
    self.df = self.query('''SELECT * FROM %t''')
  stattypes_df = self.variable_stattypes(generator_name=generator_name)
  stattypes = dict(zip(stattypes_df['name'], stattypes_df['stattype']))
  result = pandas.DataFrame()
  for desc in ('name', 'stattype',
               'unique', 'top', 'freq',
               'mean', 'std', 'min', '25%', '50%', '75%', 'max',
               'count', 'hist'):
    result[desc] = [None] * len(self.df.columns)
  for i, var in enumerate(self.df.columns):
    result['name'][i] = var
    stattype = stattypes[var] if var in stattypes else 'ignored'
    result['stattype'][i] = stattype
    for desc, value in self.df[var].describe().iteritems():
      result[desc][i] = value
  return result

@population_method(population=0)
def quick_explore_vars(self, varnames, plotfile=None, nsimilar=20):
  """Show dependence probabilities and neighborhoods based on those.

  varnames: list of strings
      At least two column names to look at dependence probabilities of,
      and to explore neighborhoods of.
  nsimilar: positive integer
      The size of the neighborhood to explore.
  """
  if len(varnames) < 2:
    raise BLE(ValueError('Need to explore at least two variables.'))
  self.pairplot_vars(varnames)
  query_columns = '''"%s"''' % '''", "'''.join(varnames)
  deps = self.query('''ESTIMATE DEPENDENCE PROBABILITY
                       FROM PAIRWISE COLUMNS OF %s
                       FOR %s;''' % (self.generator_name, query_columns))
  deps.columns = ['genid', 'name0', 'name1', 'value']
  if plotfile:
    self.logger.plot(plotfile + '-deps', self.heatmap(deps))
  deps.columns = ['genid', 'name0', 'name1', 'value']
  triangle = deps[deps['name0'] < deps['name1']]
  triangle = triangle.sort_values(ascending=False, by=['value'])
  self.logger.result("Pairwise dependence probability for: %s\n%s\n\n",
                     query_columns, triangle)

  for col in varnames:
    neighborhood = self.query(
    '''ESTIMATE *, DEPENDENCE PROBABILITY WITH "%s"
       AS "Probability of Dependence with %s"
       FROM COLUMNS OF %s
       ORDER BY "Probability of Dependence with %s"
       DESC LIMIT %d;'''
       % (col, col, self.generator_name, col, nsimilar))
    neighbor_columns = ('''"%s"''' %
                        '''", "'''.join(neighborhood["name"].tolist()))
    deps = self.query('''ESTIMATE DEPENDENCE PROBABILITY
        FROM PAIRWISE COLUMNS OF %s
        FOR %s;''' % (self.generator_name, neighbor_columns))
    deps.columns = ['genid', 'name0', 'name1', 'value']
    if plotfile:
      self.logger.plot(plotfile +"-"+ col , self.heatmap(deps))
    self.logger.result("Pairwise dependence probability of %s with its " +
                       "strongest dependents:\n%s\n\n", col, neighborhood)

@population_method(population=0)
def quick_similar_rows(self, identify_row_by, nsimilar=10):
  """Explore rows similar to the identified one.

  identify_row_by : dict
      Dictionary of column names to their values. These will be turned into
      a WHERE clause in BQL, and must identify one unique row.
  nsimilar : positive integer
      The number of similar rows to retrieve.
  """
  import hashlib
  table_name = 'tmptbl_' + hashlib.md5('\x00'.join(
      [repr(identify_row_by), str(self.status)])).hexdigest()
  column_name = 'similarity_to_' + "__".join(
      re.sub(r'\W', '_', str(val)) for val in identify_row_by.values())
  query_params = []
  query_columns = []
  for k, v in identify_row_by.iteritems():
    query_columns.append('''%s = ? ''' % bayeslite.bql_quote_name(k))
    query_params.append(v)
  query_attrs = ' and '.join(query_columns)

  with self.bdb.savepoint():
    row_exists = self.query('SELECT COUNT(*) FROM %s WHERE %s;' %
                            (self.name, query_attrs))
    if row_exists.ix[0][0] != 1:
      raise BLE(NotImplementedError(
          'identify_row_by found %d rows instead of exactly 1 in %s.' %
          (row_exists.ix[0][0], self.csv_path)))
    creation_query = ('''CREATE TEMP TABLE IF NOT EXISTS %s AS ESTIMATE *,
                         SIMILARITY TO (%s) AS %s FROM %%g LIMIT %d;''' %
                      (table_name, query_attrs, column_name, nsimilar))
    self.query(creation_query, query_params)
    result = self.query('''SELECT * FROM %s ORDER BY %s DESC;''' %
                        (table_name, column_name))
  return result



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
import pandas as pd

from sklearn.preprocessing import Imputer

def extract_sklearn_dataset(conditions, targets, dataset):
    """Extracts the `conditions` and `targets` colums from `dataset` with
    additional preprocessing.

    `NaN` strings are converted to Python `None`. Rows where the target is
    absent are dropped. All columns not in `conditions` and `target` are
    dropped.

    Parameters
    ----------
    condtions, targets : list<str>
        Column names of the `conditions` and `targets`
    dataset : pandas.DataFrame

    Returns
    -------
    pd.DataFrame
    """
    dataset = dataset.where((pd.notnull(dataset)), None)
    return dataset[conditions + targets].dropna(subset=targets)

def extract_sklearn_features_numerical(columns, dataset):
    """Extracts the numerical `columns` from `dataset`.

    Missing cells are imputed using mean imputation.

    Parameters
    ----------
    columns : list<str>
        Column names corresponding to numerical features.
    dataset : pandas.DataFrame

    Returns
    -------
    pd.DataFrame
    """
    X_numerical = dataset[columns].as_matrix().astype(float)
    return Imputer().fit_transform(X_numerical)

def extract_sklearn_features_categorical(categories,  categories_to_val_map,
        dataset):
    """Converts each categorical column i (Ki categories, N rows) into an
    N x Ki matrix. Each row in the matrix is a binary vector.

    If there are J columns in conditions_categorical, then
    dataset_binary will be N x (J*sum(Ki)) (ie all encoded categorical
    matrices are concatenated).

    Example
        Nationality|Gender
        -----------+------      -----+-
        USA        | M          1,0,0,1
        France     | F          0,1,0,0
        France     | M          0,1,0,1
        Germany    | M          0,0,1,1

    Parameters
    ----------
    categories : list<str>
        List of column names corresponding to the categoricals.
    categories_to_val_map : dict<category : dict<cat:code>>
        A dictionary of the code lookup dictionary for each category. For
        example, in the example above
        {'Nationality': {'USA':0, 'France':1, 'Germany':2},
            'Gender': {'M':0, 'F':1}}
    dataset : pandas.DataFrame
        The matrix stored as a pandas dataframe. The `categories` must appear
        as columns in the dataframe.

    Returns
    -------
    dataset_binary : np.array
        Refined dataset with the same number of rows and appropriate number
        of columns representing the binary version of dataset.
    """
    dataset_binary = []
    for row in dataset.iterrows():
        row = list(row[1][categories])
        row_binary = binarize_categorical_row(
            categories, categories_to_val_map, row)
        dataset_binary.append(row_binary)
    return np.asarray(dataset_binary)

def binarize_categorical_row(categories, categories_to_val_map, row):
    """Unrolls a row of categorical data into the corresponding binary
    vector version. The order of the entries in `row` must be the same as those
    in the list `categories`. The `row` must be a list of strings corresponding
    to the value of each categorical column.
    """
    assert len(row) == len(categories)
    binary_data = []
    for categorical, value in zip(categories, row):
        K = len(categories_to_val_map[categorical])
        encoding = [0]*K
        encoding[categories_to_val_map[categorical][value]] = 1
        binary_data.extend(encoding)
    return binary_data

def build_categorical_to_value_map(columns, dataset):
    """Builds a dictionary of dictionaries.

    Parameters
    ----------
    columns : list<str>
        Column names corresponding to categorical features.
    dataset : pandas.DataFrame

    Returns
    -------
    categories_to_val_map : dict<col:dict>
        Dictionary of keys with dictionary values. Each value dictionary
        contains the mapping category -> code for the corresponding
        categorical feature, which is the key.
    """
    categories_to_val_map = dict()
    for categorical in columns:
        categories_to_val_map[categorical] = {val:code
            for (code,val) in enumerate(dataset[categorical].unique())}
    return categories_to_val_map

def extract_sklearn_univariate_target(target, dataset):
    """Extracts a single target column from a dataset as a vector for sklearn.

    Parameters
    ----------
    target : str
        Name of the target column.
    dataset : pandas.DataFrame

    Returns
    -------
    target_vector : np.array
    """
    return dataset[target].as_matrix().ravel()



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pandas as pd

import bayeslite.core
from bayeslite import bayesdb_open
from bayeslite import bql_quote_name as quote
from bayeslite.exception import BayesLiteException as BLE
from bayeslite.loggers import logged_query
from bayeslite.read_pandas import bayesdb_read_pandas_df
from bayeslite.sqlite3_util import sqlite3_quote_name
from bayeslite.util import cursor_value
from bdbcontrib.population_method import population_method

from bdbcontrib.population_method import population_method

###############################################################################
###                                 PUBLIC                                  ###
###############################################################################

@population_method(population_to_bdb=0, population_name=1)
def cardinality(bdb, table, cols=None):
    """Compute the number of unique values in the columns of a table.

    Parameters
    ----------
    bdb : __population_to_bdb__
    table : __population_name__
        Name of table.
    cols : list<str>, optional
        Columns to compute the unique values. Defaults to all.

    Returns
    -------
    counts : pandas.DataFrame whose .columns are ['name', 'distinct_count'].
    """
    # If no columns specified, use all.
    if not cols:
        sql = 'PRAGMA table_info(%s)' % (quote(table),)
        res = bdb.sql_execute(sql)
        cols = [r[1] for r in res]

    names=[]
    counts=[]
    for col in cols:
        sql = '''
            SELECT COUNT (DISTINCT %s) FROM %s
        ''' % (quote(col), quote(table))
        res = bdb.sql_execute(sql)
        names.append(col)
        counts.append(cursor_value(res))
    return pd.DataFrame({'name': names, 'distinct_count': counts})


@population_method(population_to_bdb=0, population_name=1)
def nullify(bdb, table, value):
    """Replace specified values in a SQL table with ``NULL``.

    Parameters
    ----------
    bdb : __population_to_bdb__
    table : str
        The name of the table on which to act
    value : stringable
        The value to replace with ``NULL``

    Examples
    --------
    >>> import bayeslite
    >>> from bdbcontrib import plotutils
    >>> with bayeslite.bayesdb_open('mydb.bdb') as bdb:
    >>>    bdbcontrib.nullify(bdb, 'mytable', 'NaN')
    """
    # get a list of columns of the table
    c = bdb.sql_execute('pragma table_info({})'.format(quote(table)))
    columns = [r[1] for r in c]
    for col in columns:
        if value in ["''", '""']:
            bql = '''
                UPDATE {} SET {} = NULL WHERE {} = '';
            '''.format(quote(table), quote(col), quote(col))
            bdb.sql_execute(bql)
        else:
            bql = '''
                UPDATE {} SET {} = NULL WHERE {} = ?;
            '''.format(quote(table), quote(col), quote(col))
            bdb.sql_execute(bql, (value,))


def cursor_to_df(cursor):
    """Converts SQLite3 cursor to a pandas DataFrame."""
    # Do this in a savepoint to enable caching from row to row in BQL
    # queries.
    with cursor.connection.savepoint():
        df = pd.DataFrame.from_records(cursor, coerce_float=True)
    if not df.empty:
        df.columns = [desc[0] for desc in cursor.description]
        for col in df.columns:
            try:
                df[col] = df[col].astype(float)
            except ValueError:
                pass

    return df

def table_to_df(bdb, table_name, column_names=None):
    """Return the contents of the given table as a pandas DataFrame.

    If `column_names` is not None, fetch only those columns.
    """
    qt = sqlite3_quote_name(table_name)
    if column_names is not None:
        qcns = ','.join(map(sqlite3_quote_name, column_names))
        select_sql = 'SELECT %s FROM %s' % (qcns, qt)
    else:
        select_sql = 'SELECT * FROM %s' % (qt,)
    return cursor_to_df(bdb.sql_execute(select_sql))

def df_to_table(df, tablename=None, **kwargs):
    """Return a new BayesDB with a single table with the data in `df`.

    `df` is a Pandas DataFrame.

    If `tablename` is not supplied, an arbitrary one will be chosen.

    `kwargs` are passed on to `bayesdb_open`.

    Returns a 2-tuple of the new BayesDB instance and the name of the
    new table.
    """
    bdb = bayesdb_open(**kwargs)
    if tablename is None:
        tablename = bdb.temp_table_name()
    bayesdb_read_pandas_df(bdb, tablename, df, create=True)
    return (bdb, tablename)

@population_method(population_to_bdb=0, interpret_bql=1, logger="logger")
def query(bdb, bql, bindings=None, logger=None):
    """Execute the `bql` query on the `bdb` instance.

    Parameters
    ----------
    bdb : __population_to_bdb__
    bql : __interpret_bql__
    bindings : Values to safely fill in for '?' in the BQL query.

    Returns
    -------
    df : pandas.DataFrame
        Table of results as a pandas dataframe.
    """
    if bindings is None:
        bindings = ()
    if logger:
        logger.info("BQL [%s] %s", bql, bindings)
    cursor = bdb.execute(bql, bindings)
    return cursor_to_df(cursor)

@population_method(population_to_bdb=0, population_name=1)
def describe_table(bdb, table_name):
    """Returns a DataFrame containing description of `table_name`.

    Examples
    --------
    >>> bdbcontrib.describe_table(bdb, 'employees')
    tabname   | colno |    name
    ----------+-------+--------
    employees |     0 |    name
    employees |     1 |     age
    employees |     2 |  weight
    employees |     3 |  height
    """
    if not bayeslite.core.bayesdb_has_table(bdb, table_name):
            raise BLE(NameError('No such table {}'.format(table_name)))
    sql = '''
        SELECT tabname, colno, name
            FROM bayesdb_column
            WHERE tabname=?
            ORDER BY tabname ASC, colno ASC
        '''
    curs = bdb.sql_execute(sql, bindings=(table_name,))
    return cursor_to_df(curs)


@population_method(population_to_bdb=0, generator_name=1)
def describe_generator(bdb, generator_name):
    """Returns a DataFrame containing description of `generator_name`.

    Examples
    --------

    >>> bdbcontrib.describe_generator(bdb, 'employees_gen')
    id |          name |   tabname | metamodel
    ---+---------------+-----------+----------
    3  | employees_gen | employees |  crosscat
    """
    if not bayeslite.core.bayesdb_has_generator_default(bdb, generator_name):
            raise BLE(NameError('No such generator {}'.format(generator_name)))
    sql = '''
            SELECT id, name, tabname, metamodel
                FROM bayesdb_generator
                WHERE name = ?
        '''
    curs = bdb.sql_execute(sql, bindings=(generator_name,))
    return cursor_to_df(curs)

@population_method(population_to_bdb=0, generator_name='generator_name')
def variable_stattypes(bdb, generator_name=None):
    assert generator_name
    """The modeled statistical types of each variable in order."""
    if not bayeslite.core.bayesdb_has_generator_default(bdb, generator_name):
            raise BLE(NameError('No such generator {}'.format(generator_name)))
    sql = '''
        SELECT c.colno AS colno, c.name AS name,
                gc.stattype AS stattype
            FROM bayesdb_generator AS g,
                (bayesdb_column AS c LEFT OUTER JOIN
                    bayesdb_generator_column AS gc
                    USING (colno))
            WHERE g.id = ? AND g.id = gc.generator_id
                AND g.tabname = c.tabname
            ORDER BY colno ASC;
    '''
    generator_id = bayeslite.core.bayesdb_get_generator_default(bdb,
        generator_name)
    curs = bdb.sql_execute(sql, bindings=(generator_id,))
    return cursor_to_df(curs)

@population_method(population_to_bdb=0)
def list_metamodels(bdb):
    df = query(bdb, "SELECT name FROM bayesdb_generator;")
    return list(df['name'])

@population_method(population_to_bdb=0)
def list_tables(bdb):
    df = query(bdb, """SELECT name FROM sqlite_master
                       WHERE type='table' AND
                           NAME NOT LIKE "bayesdb_%" AND
                           NAME NOT LIKE "sqlite_%";""")
    return list(df['name'])

@population_method(population_to_bdb=0, generator_name=1)
def describe_generator_models(bdb, generator_name):
    """Returns a DataFrame containing description of the models
    in `generator_name`.

    Examples
    --------

    >>> bdbcontrib.describe_generator_models(bdb, 'employees_gen')
    modelno | iterations
    --------+-----------
          0 | 100
    """
    if not bayeslite.core.bayesdb_has_generator_default(bdb, generator_name):
            raise BLE(NameError('No such generator {}'.format(generator_name)))
    sql = '''
        SELECT modelno, iterations FROM bayesdb_generator_model
            WHERE generator_id = ?
        '''
    generator_id = bayeslite.core.bayesdb_get_generator_default(bdb,
        generator_name)
    curs = bdb.sql_execute(sql, bindings=(generator_id,))
    return cursor_to_df(curs)


###############################################################################
###                              INTERNAL                                   ###
###############################################################################

def get_column_info(bdb, generator_name):
    generator_id = bayeslite.core.bayesdb_get_generator(bdb, generator_name)
    sql = '''
        SELECT c.colno, c.name, gc.stattype
            FROM bayesdb_generator AS g,
                bayesdb_generator_column AS gc,
                bayesdb_column AS c
            WHERE g.id = ?
                AND gc.generator_id = g.id
                AND gc.colno = c.colno
                AND c.tabname = g.tabname
            ORDER BY c.colno
    '''
    return bdb.sql_execute(sql, (generator_id,)).fetchall()

@population_method(population_to_bdb=0, generator_name=1)
def get_column_stattype(bdb, generator_name, column_name):
    generator_id = bayeslite.core.bayesdb_get_generator(bdb, generator_name)
    sql = '''
        SELECT gc.stattype
            FROM bayesdb_generator AS g,
                bayesdb_generator_column AS gc,
                bayesdb_column AS c
            WHERE g.id = ?
                AND gc.generator_id = g.id
                AND gc.colno = c.colno
                AND c.name = ?
                AND c.tabname = g.tabname
            ORDER BY c.colno
    '''
    cursor = bdb.sql_execute(sql, (generator_id, column_name,))
    try:
        row = cursor.next()
    except StopIteration:
        # XXX Temporary kludge for broken callers.
        raise IndexError
    else:
        return row[0]

@population_method(population=0, generator_name='generator_name')
def analyze(self, models=100, minutes=0, iterations=0, checkpoint=0,
            generator_name=None):
  '''Run analysis.

  models : integer
      The number of models bounds the accuracy of predictive probabilities.
      With ten models, then you get one decimal digit of interpretability,
      with a hundred models, you get two, and so on.
  minutes : integer
      How long you want to let it run.
  iterations : integer
      How many iterations to let it run.

  Returns:
      A report indicating how many models have seen how many iterations,
      and other info about model stability.
  '''
  assert generator_name is not None
  if models > 0:
    self.query('INITIALIZE %d MODELS IF NOT EXISTS FOR %s' %
          (models, generator_name))
    assert minutes == 0 or iterations == 0
  else:
    models = self.analysis_status(generator_name=generator_name).sum()
  if minutes > 0:
    if checkpoint == 0:
      checkpoint = max(1, int(minutes * models / 200))
      analyzer = ('ANALYZE %s FOR %d MINUTES CHECKPOINT %d ITERATION WAIT' %
                  (generator_name, minutes, checkpoint))
      with logged_query(query_string=analyzer,
                        name=self.session_capture_name,
                        bindings=self.query('SELECT * FROM %t')):
        self.query(analyzer)
  elif iterations > 0:
    if checkpoint == 0:
      checkpoint = max(1, int(iterations / 20))
    self.query(
        '''ANALYZE %s FOR %d ITERATIONS CHECKPOINT %d ITERATION WAIT''' % (
            generator_name, iterations, checkpoint))
  else:
    raise NotImplementedError('No default analysis strategy yet. '
                              'Please specify minutes or iterations.')
  # itrs = self.per_model_analysis_status()
  # models_with_fewest_iterations =
  #    itrs[itrs['iterations'] == itrs.min('index').head(0)[0]].index.tolist()
  # TODO(gremio): run each model with as many iterations as it needs to get
  # up to where it needs to get to, if that's larger?
  # Nope. Vikash said there's no reason to think that's a good idea. Perhaps
  # even better to have some young models mixed in with the old ones.
  # I still think we should make some recommendation that scales for what
  # "the right thing" is, where that's something that at least isn't known to
  # suck.

  return self.analysis_status(generator_name=generator_name)

@population_method(population=0, generator_name='generator_name')
def per_model_analysis_status(self, generator_name=None):
  """Return the number of iterations for each model."""
  assert generator_name is not None
  try:
    return self.query('''SELECT iterations FROM bayesdb_generator_model
                         WHERE generator_id = (
                          SELECT id FROM bayesdb_generator WHERE name = ?)''',
                      (generator_name,))
  except ValueError:
    # Because, e.g. there is no generator yet, for an empty db.
    return None

@population_method(population=0, generator_name='generator_name')
def analysis_status(self, generator_name=None):
  """Return the count of models for each number of iterations run."""
  assert generator_name is not None
  itrs = self.per_model_analysis_status(generator_name=generator_name)
  if itrs is None or len(itrs) == 0:
    emt = pd.DataFrame(columns=['count of model instances'])
    emt.index.name = 'iterations'
    return emt
  vcs = pd.DataFrame(itrs['iterations'].value_counts())
  vcs.index.name = 'iterations'
  vcs.columns = ['count of model instances']
  self.status = vcs
  return vcs

def get_data_as_list(bdb, table_name, column_list=None):
    if column_list is None:
        sql = '''
            SELECT * FROM {};
        '''.format(quote(table_name))
    else:
        sql = '''
            SELECT {} FROM {}
        '''.format(', '.join(map(quote, column_list)), table_name)
    cursor = bdb.sql_execute(sql)
    T = cursor_to_df(cursor).values.tolist()
    return T


def get_shortnames(bdb, table_name, column_names):
    return get_column_descriptive_metadata(bdb, table_name, column_names,
        'shortname')


def get_descriptions(bdb, table_name, column_names):
    return get_column_descriptive_metadata(bdb, table_name, column_names,
        'description')


def get_column_descriptive_metadata(bdb, table_name, column_names, md_field):
    short_names = []
    # XXX: this is indefensibly wasteful.
    bql = '''
        SELECT colno, name, {} FROM bayesdb_column WHERE tabname = ?
    '''.format(md_field)
    records = bdb.sql_execute(bql, (table_name,)).fetchall()

    # hack for case sensitivity problems
    column_names = [c.upper().lower() for c in column_names]
    records = [(r[0], r[1].upper().lower(), r[2]) for r in records]

    for cname in column_names:
        for record in records:
            if record[1] == cname:
                sname = record[2]
                if sname is None:
                    sname = cname
                short_names.append(sname)
                break

    assert len(short_names) == len(column_names)
    return short_names



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
import pandas as pd

from bdbcontrib.bql_utils import df_to_table
from crosscat.tests import synthetic_data_generator as sdg

from bdbcontrib.predictors.random_forest import RandomForest
from bdbcontrib.predictors.keplers_law import KeplersLaw
from bdbcontrib.predictors.multiple_regression import MultipleRegression

# TODO: More robust tests exploring more interesting cases. The main use
# right now is crash testing. Moreover common patterns can be automated.

# ------------------------------------------------------------------------------
# Helper

def get_synthetic_data(n_sample, seed=438):
    cols_to_views = [
        0, 0, 0, 0, 0,
        1, 1, 1, 1,
        2, 2, 2,
        3,
        4
    ]
    colnames = [
        'm1', 'c1', 'm2', 'c2', 'm3',
        'm4', 'c3', 'c4', 'm5',
        'c5', 'c6', 'c7',
        'c8',
        'c9'
    ]
    cctypes = [
        'multinomial', 'continuous', 'multinomial', 'continuous', 'multinomial',
        'multinomial', 'continuous', 'continuous', 'multinomial',
        'continuous', 'continuous', 'continuous',
        'continuous',
        'continuous'
    ]
    distargs = [
        dict(K=9), None, dict(K=9), None, dict(K=7),
        dict(K=4), None, None, dict(K=9),
        None, None, None,
        None,
        None
    ]
    component_weights = [
        [.2, .3, .5],
        [.9, .1],
        [.4, .4, .2],
        [.8, .2],
        [.4, .5, .1]
    ]
    separation = [0.8, 0.9, 0.65, 0.7, 0.75]
    synthetic_data = sdg.gen_data(cctypes, n_sample, cols_to_views,
        component_weights, separation, seed, distargs=distargs)
    data = pd.DataFrame(synthetic_data[0])
    data.columns = colnames
    return df_to_table(data)

# ------------------------------------------------------------------------------
# Tests

def test_random_forest():
    # Train foreign predictor.
    (bdb, table) = get_synthetic_data(150)
    conditions = [(c, 'NUMERICAL') for c in ['c1','c2','c4','c8']] + \
        [(c, 'CATEGORICAL') for c in ['m1', 'm3']]
    target = [('m5', 'CATEGORICAL')]
    srf_predictor = RandomForest.create(bdb, table, target, conditions)

    # Dummy realization of conditions. Include extra conditions with unseen.
    parents = {'c1':1.3, 'c2':-2.1, 'c4':0.2, 'c8':0.2, 'm1':1, 'm3':7, 'm5':5}

    # Crash tests.
    srf_predictor.simulate(10, parents)
    pdf_val = srf_predictor.logpdf(7, parents)
    assert srf_predictor.logpdf(-1, parents) == float('-inf')

    # Dummy realization of conditions. Include extra conditions all seen.
    parents = {'c1':1.3, 'c2':-2.1, 'c4':0.2, 'c8':0.2, 'm1':1, 'm3':4, 'm5':5}

    # Crash tests.
    srf_predictor.simulate(10, parents)
    pdf_val = srf_predictor.logpdf(7, parents)
    assert srf_predictor.logpdf(-1, parents) == float('-inf')

    # Serialization tests.
    srf_binary = RandomForest.serialize(bdb, srf_predictor)
    srf_predictor2 = RandomForest.deserialize(bdb, srf_binary)
    srf_predictor2.simulate(10, parents)
    pdf_val2 = srf_predictor2.logpdf(7, parents)
    assert np.allclose(pdf_val, pdf_val2)

def test_keplers_law():
    # Train foreign predictor.
    (bdb, table) = get_synthetic_data(150)
    conditions = [('c1','NUMERICAL'), ('c3', 'NUMERICAL')]
    target = [('c4', 'NUMERICAL')]
    kl_predictor = KeplersLaw.create(bdb, table, target, conditions)

    # Dummy realization of conditions.
    inputs = {'c1':2.1, 'c3':1.7}

    # Crash tests.
    kl_predictor.simulate(10, inputs)
    pdf_val = kl_predictor.logpdf(1.2, inputs)

    # Serialization tests.
    kl_binary = KeplersLaw.serialize(bdb, kl_predictor)
    kl_predictor2 = KeplersLaw.deserialize(bdb, kl_binary)
    kl_predictor2.simulate(10, inputs)
    pdf_val2 = kl_predictor2.logpdf(1.2, inputs)
    assert np.allclose(pdf_val, pdf_val2)

def test_multiple_regression():
    # Train foreign predictor.
    (bdb, table) = get_synthetic_data(150)
    conditions = [(c, 'NUMERICAL') for c in ['c1','c2','c4','c8']] + \
        [(c, 'CATEGORICAL') for c in ['m1', 'm3']]
    target = [('c7', 'NUMERICAL')]
    mr_predictor = MultipleRegression.create(bdb, table, target, conditions)

    # Dummy realization of conditions. Include extra conditions with unseen.
    inputs = {'c1':1.3, 'c2':-2.1, 'c4':0.2, 'c8':0.2, 'm1':1, 'm3':7, 'm5':5}

    # Crash tests.
    mr_predictor.simulate(10, inputs)
    pdf_val = mr_predictor.logpdf(-0.4, inputs)

    # Dummy realization of conditions. Include extra conditions al seen.
    inputs = {'c1':1.3, 'c2':-2.1, 'c4':0.2, 'c8':0.2, 'm1':1, 'm3':4, 'm5':5}

    # Crash tests.
    mr_predictor.simulate(10, inputs)
    pdf_val = mr_predictor.logpdf(-0.4, inputs)

    # Serialization tests.
    mr_binary = MultipleRegression.serialize(bdb, mr_predictor)
    mr_predictor2 = MultipleRegression.deserialize(bdb, mr_binary)
    mr_predictor2.simulate(10, inputs)
    pdf_val2 = mr_predictor2.logpdf(-0.4, inputs)
    assert np.allclose(pdf_val, pdf_val2)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import sys

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

from io import BytesIO
import mock
import multiprocessing
import numpy as np
import re
import os
import pandas as pd
import pytest
import random
import shutil
from string import ascii_lowercase  # pylint: disable=deprecated-module
import tempfile

import bayeslite
import bdbcontrib

from bayeslite.exception import BayesLiteException as BLE
from bayeslite.loggers import CaptureLogger
from bayeslite.read_csv import bayesdb_read_csv
from bdbcontrib.bql_utils import cursor_to_df
from bdbcontrib.plot_utils import _pairplot

def dataset(num_rows=400):
    '''Column names give rough type, numbers show which are related.'''
    df = pd.DataFrame()
    alphabet = ['A', 'B', 'C', 'D', 'E']
    dist_of_five = np.random.choice(range(5), num_rows,
                                    p=np.array([1, .4, .3, .2, .1]) / 2.)
    df['floats_1'] = [np.random.randn() + x for x in dist_of_five]
    df['categorical_1'] = [alphabet[i] for i in dist_of_five]

    dist_of_five = np.random.choice(range(5), num_rows,
                                    p=np.array([1, .4, .3, .2, .1]) / 2.)
    df['categorical_2'] = [alphabet[i] for i in dist_of_five]

    few_ints = np.random.choice(range(4), num_rows, p=[.4, .3, .2, .1])
    df['few_ints_3'] = few_ints
    df['floats_3'] = [(np.random.randn() - 2 * x) / (1 + x) for x in few_ints]

    df['many_ints_4'] = np.random.randn(num_rows)

    # Need >= 75% to be zero to trigger inter-quartile-range == 0.
    numeric_5 = np.random.choice(
        range(4), num_rows, p=[200./256, 40./256, 10./256, 6./256])
    # Need >= 30 unique to be numeric instead of categorical,
    # so perturb the non-zeros:
    df['skewed_numeric_5'] = [ i * np.random.random() for i in numeric_5 ]
    csv_data = BytesIO()
    df.to_csv(csv_data, header=True, index_label='index')
    return (df, csv_data)

def ensure_timeout(delay, target):
    proc = multiprocessing.Process(target=target)
    proc.start()
    proc.join(delay)
    assert not proc.is_alive()

# Session scope (entire test run) rather than module scope (this test) because
# it will be re-used by other test files testing other modules:
@pytest.fixture(scope='session')
def dts_df(request):
    (df, csv_data) = dataset(40)
    tempd = tempfile.mkdtemp(prefix="bdbcontrib-test-recipes")
    request.addfinalizer(lambda: shutil.rmtree(tempd))
    csv_path = os.path.join(tempd, "data.csv")
    with open(csv_path, "w") as csv_f:
        csv_f.write(csv_data.getvalue())
    bdb_path = os.path.join(tempd, "data.bdb")
    name = ''.join(random.choice(ascii_lowercase) for _ in range(32))
    dts = bdbcontrib.Population(name=name, csv_path=csv_path, bdb_path=bdb_path,
        logger=CaptureLogger(verbose=pytest.config.option.verbose),
        session_capture_name="test_recipes.py")
    ensure_timeout(10, lambda: dts.analyze(models=10, iterations=20))
    return dts, df

def prepare():
    (df, csv_str) = dataset()
    os.environ['BAYESDB_WIZARD_MODE']='1'
    bdb = bayeslite.bayesdb_open()
    # XXX Do we not have a bayesdb_read_df ?
    bayesdb_read_csv(bdb, 'plottest', flush(csv_str),
                     header=True, create=True)
    bdb.execute('''
        create generator plottest_cc for plottest using crosscat(guess(*))
    ''')

    # do a plot where a some sub-violins are removed
    _remove_violin_bql = """
        DELETE FROM plottest
            WHERE categorical_1 = "B"
                AND (few_ints_3 = 2 OR few_ints_3 = 1);
    """
    cursor = bdb.execute('SELECT * FROM plottest')
    df = cursor_to_df(cursor)
    return (df, bdb)

def run_pairplot(prepped, location, **kwargs):
    (df, bdb) = prepped
    plt.figure(tight_layout=True, facecolor='white')
    _pairplot(df, bdb=bdb, generator_name='plottest_cc',
              show_full=False, **kwargs)
    plt.savefig(location)
    print 'wrote pairplot to %r' % (location,)

def run_histogram(bdb, df, location, **kwargs):
    plt.figure(tight_layout=True, facecolor='white')
    bdbcontrib.plot_utils.histogram(bdb, df)
    plt.savefig(location)
    print 'wrote histogram to %r' % (location,)

def run_mi_hist(bdb, location, gen, col1, col2, *args, **kwargs):
    plt.figure()
    bdbcontrib.plot_utils.mi_hist(bdb, gen, col1, col2, *args, **kwargs)
    plt.savefig(location)
    print 'wrote mi_hist to %r' % (location,)

def run_heatmap(bdb, location, gen, *args, **kwargs):
    plt.figure()
    qg = bayeslite.bql_quote_name(gen)
    df = cursor_to_df(bdb.execute('''
        estimate dependence probability from pairwise columns of %s
    ''' % (qg,)))
    bdbcontrib.plot_utils.heatmap(df, *args, **kwargs)
    plt.show()
    plt.savefig(location)
    print 'wrote heatmap to %r' % (location,)

from PIL import Image
def has_nontrivial_contents_over_white_background(imgfile):
    img = Image.open(imgfile)
    pix = img.load()
    hist = img.histogram()
    # White background means corners are likely white:
    assert pix[0, 0] == (255, 255, 255, 255)
    assert pix[img.size[0] - 1, img.size[1] - 1] == (255, 255, 255, 255)
    white = hist[-1]
    total = sum(hist)
    # Total white should be between 10 and 90% of the image.
    assert total * .1 < white
    assert white < total * .9
    # Total number of different intensities used should be between 10 and 90%
    # of the available intensities.
    zeros = len([intensity for intensity in hist if intensity == 0])
    assert len(hist) * .1 < zeros
    assert zeros < len(hist) * .9
    return True

def flush(f):
    return BytesIO(f.getvalue())

def test_pairplot_smoke():
    ans = prepare()
    f = BytesIO()
    run_pairplot(ans, f, colorby='categorical_2', show_contour=False)
    assert has_nontrivial_contents_over_white_background(flush(f))
    f = BytesIO()
    run_pairplot(ans, f, colorby='categorical_2', show_contour=True)
    assert has_nontrivial_contents_over_white_background(flush(f))
    f = BytesIO()
    run_pairplot(ans, f, show_contour=False)
    assert has_nontrivial_contents_over_white_background(flush(f))

def test_mi_hist_smoke():
    df, bdb = prepare()
    bdb.execute('initialize 10 models for plottest_cc')
    f = BytesIO()
    run_mi_hist(bdb, f, 'plottest_cc', 'floats_1', 'categorical_1',
        num_samples=10, bins=4)
    f = BytesIO()
    run_mi_hist(bdb, f, 'plottest_cc', 'few_ints_3', 'many_ints_4')

def test_heatmap_smoke():
    df, bdb = prepare()
    bdb.execute('initialize 10 models for plottest_cc')
    f = BytesIO()
    run_heatmap(bdb, f, 'plottest_cc')

def test_one_variable():
    (df, bdb) = prepare()
    for var in ['categorical_1', 'few_ints_3', 'floats_3', 'many_ints_4',
                'skewed_numeric_5']:
      cursor = bdb.execute('SELECT %s FROM plottest' % (var,))
      df = cursor_to_df(cursor)
      f = BytesIO()
      run_pairplot((df, bdb), f, show_contour=False)
      assert has_nontrivial_contents_over_white_background(flush(f))
      cursor = bdb.execute('SELECT %s, categorical_2 FROM plottest' % (var,))
      df = cursor_to_df(cursor)
      f = BytesIO()
      run_pairplot((df, bdb), f, colorby='categorical_2', show_contour=False)
      assert has_nontrivial_contents_over_white_background(flush(f))
      f = BytesIO()
      run_pairplot((df, bdb), f, colorby='categorical_2', show_contour=True)
      assert has_nontrivial_contents_over_white_background(flush(f))
    with pytest.raises(BLE) as exc:
      run_pairplot((df, bdb), f, colorby='floats_3')
      assert 'non-categorical' in str(exc)
    with pytest.raises(BLE):
      run_pairplot((df, bdb), f, colorby='categorical_2',
                   stattypes={'categorical_2': 'numerical'})
      assert 'non-categorical' in str(exc)


def test_complete_the_square():
    df = pd.DataFrame([['arb', 'arb', 1],
                       ['arb', 'bar', 0],
                       ['arb', 'caz', .2],
#                       ['bar', 'arb', 0],   # Missing.
                       ['bar', 'bar', 1],
                       ['bar', 'caz', .5],
                       ['bar', 'caz', 10000],    # Duplicate, diff value.
                       ['caz', 'arb', .2],
                       ['caz', 'bar', .5],
                       ['caz', 'caz', 1]])
    df.columns = ['i', 'c', 'v']
    xp = pd.DataFrame([['arb', 'arb', 1],
                       ['arb', 'bar', 0],
                       ['arb', 'caz', .2],
                       ['bar', 'arb', 0],    # Missing filled.
                       ['bar', 'bar', 1],
                       ['bar', 'caz', .5],   # Chose first one.
                       ['caz', 'arb', .2],
                       ['caz', 'bar', .5],
                       ['caz', 'caz', 1]])
    xp.columns = ['i', 'c', 'v']

    ob = bdbcontrib.plot_utils.ensure_full_square(
        df, pivot_kws={'index': df.columns[0],
                       'columns': df.columns[1],
                       'values': df.columns[2]})
    assert str(xp) == str(ob)

def test_select_heatmap(dts_df):
    dts, _df = dts_df
    plot = dts.heatmap('''SELECT categorical_1, categorical_2,
                          COUNT(categorical_2) AS cat2_count
                          FROM %t
                          GROUP BY categorical_2
                          HAVING COUNT(categorical_2) >= 1;''')
    f = BytesIO()
    plot.savefig(f)
    assert has_nontrivial_contents_over_white_background(flush(f))


def test_depprob_heatmap(dts_df):
    dts, _df = dts_df
    tiny_plot = dts.heatmap('ESTIMATE DEPENDENCE PROBABILITY'
                            ' FROM PAIRWISE COLUMNS OF %g'
                            ' WHERE name0 LIKE "categorical%"')
    f = BytesIO()
    tiny_plot.savefig(f)
    assert has_nontrivial_contents_over_white_background(flush(f))
    tiny_rows = tuple(tiny_plot.dendrogram_row.reordered_ind)
    tiny_cols = tuple(tiny_plot.dendrogram_col.reordered_ind)
    plt.figure(1)

    plot = dts.heatmap('ESTIMATE DEPENDENCE PROBABILITY'
                       ' FROM PAIRWISE COLUMNS OF %g')
    rows = plot.dendrogram_row.reordered_ind
    cols = plot.dendrogram_col.reordered_ind
    assert len(tiny_rows) < len(rows)
    assert len(tiny_cols) == len(cols) # Note: only restricted name0
    assert sorted(rows) != rows
    assert sorted(cols) != cols

    (sorted_plot, srows, scols) = dts.heatmap('ESTIMATE DEPENDENCE PROBABILITY'
        ' FROM PAIRWISE COLUMNS OF %g',
        row_ordering=sorted(plot.dendrogram_row.reordered_ind),
        col_ordering=sorted(plot.dendrogram_col.reordered_ind))

    assert rows != srows
    assert cols != scols
    assert sorted(rows) == srows
    assert sorted(cols) == scols


def test_histogram():
    categoricals = set(['categorical_1', 'categorical_2', 'few_ints_3'])
    numerics = set(['floats_1', 'floats_3', 'many_ints_4', 'skewed_numeric_5'])
    (df, bdb) = prepare()
    f = BytesIO()
    with pytest.raises(BLE):
        run_histogram(bdb, pd.DataFrame(), f)
    with pytest.raises(BLE):
        run_histogram(bdb, df[['categorical_1', 'few_ints_3', 'floats_3']], f)
    for datacol in categoricals:
        with pytest.raises(BLE):
            run_histogram(bdb, df[[datacol]], f)
        with pytest.raises(BLE):
            run_histogram(bdb, df[[datacol, 'categorical_1']], f)
        with pytest.raises(BLE):
            run_histogram(bdb, df[[datacol, 'many_ints_4']], f)
    for colorby in numerics:
        with pytest.raises(BLE):
            run_histogram(bdb, df[['many_ints_4', colorby]], f)

    for datacol in numerics:
        for colorby in categoricals:
            f = BytesIO()
            run_histogram(bdb, df[[datacol]], f)
            assert has_nontrivial_contents_over_white_background(flush(f))
            if datacol == colorby:
                with pytest.raises(BLE):
                    f = BytesIO()
                    run_histogram(bdb, df[[datacol, colorby]], f)
            else:
                f = BytesIO()
                run_histogram(bdb, df[[datacol, colorby]], f)
                assert has_nontrivial_contents_over_white_background(flush(f))

def get_plot_text(obj):
    texts = []
    stack = [obj]
    while stack:
        item = stack.pop()
        d = set(dir(item))
        if 'get_children' in d:
            stack += item.get_children()
        if 'get_text' in d:
            texts.append(item.get_text())
    return texts

def test_gen_collapsed_legend_from_dict():
    hl_colors_dict = {'roses': 'red',
                      'violets': 'blue',
                      'lilies': 'white',
                      'poppies': 'red'}
    # http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.legend
    loc = 10  # center
    title = 'Doggerel!'
    fontsize = 'medium'
    leg = bdbcontrib.plot_utils.gen_collapsed_legend_from_dict(
        hl_colors_dict, loc, title, fontsize)
    texts = get_plot_text(leg)
    assert 'Doggerel!' in texts
    assert 'violets' in texts
    assert 'lilies' in texts
    assert ('roses, poppies' in texts or 'poppies, roses' in texts)


def main():
    ans = prepare()
    run_pairplot(ans, 'pp0.png', colorby='categorical_2', show_contour=False)
    run_pairplot(ans, 'pp1.png', colorby='categorical_2', show_contour=True)
    run_pairplot(ans, 'pp2.png', show_contour=False)
    assert os.path.exists('pp0.png')
    assert os.path.exists('pp1.png')
    assert os.path.exists('pp2.png')
    df, bdb = ans
    bdb.execute('initialize 100 models for plottest_cc')
    bdb.execute('analyze plottest_cc for 2 iterations wait')
    run_mi_hist(bdb, 'mi0.png', 'plottest_cc', 'floats_1', 'categorical_1',
        num_samples=100, bins=5)
    run_mi_hist(bdb, 'mi1.png', 'plottest_cc', 'few_ints_3', 'many_ints_4',
        num_samples=1000, bins=10)
    assert os.path.exists('mi0.png')
    assert os.path.exists('mi1.png')
    run_heatmap(bdb, 'depprob.png', 'plottest_cc')
    assert os.path.exists('depprob.png')

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
import pandas as pd
from bdbcontrib.predictors import sklearn_utils as sku

def test_extract_sklearn_dataset():
    dataset = pd.DataFrame({
        'A':[1.1, 2.1, 3.9, 4.5, 5.1],
        'B':[5.1, 4.1, 3.9, 2.5, 1.1],
        'C':['1', '2', '3', '4', '5'],
        'D':[1, None, 3, 4, 5],
        })
    conditions, targets = ['A', 'B'], ['D']
    df = sku.extract_sklearn_dataset(conditions, targets, dataset)
    # Test that column 'C' is not included.
    assert set(df.columns) == set(conditions + targets)
    # Test that the second row is dropped (it has a None target).
    assert len(df) == 4

def test_extract_sklearn_features_numerical():
    dataset = pd.DataFrame({
        'A':[1, 2, None, 4, 10],
        'B':[0, -2, 3, None, 1],
        'C':['1', '2', '3', '4', '5'],
        'D':[1, None, 3, 4, 5],
        })
    columns = ['A', 'B']
    matrix = sku.extract_sklearn_features_numerical(columns, dataset)
    # Test correct imputation.
    mean_A = (1 + 2 + 4 + 10) / 4.
    mean_B = (0 -2 + 3 + 1) / 4.
    expected = np.asarray([
        [1, 2, mean_A, 4, 10],
        [0, -2, 3, mean_B, 1]]).T
    assert np.array_equal(matrix, expected)

def test_extract_sklearn_features_categorical():
    dataset = pd.DataFrame({
        'Nationality':['USA', 'USA', 'France', 'Germany', 'Bengal'],
        'Gender':['M', 'F', 'M', 'M', 'T'],
        'C':['1', '2', '3', '4', '5'],
        'D':[1, None, 3, 4, 5],
        })
    categories = ['Nationality', 'Gender']
    categories_to_val_map = {
        'Nationality' : {'USA':0, 'France':1, 'Germany': 2, 'Bengal':3},
        'Gender' : {'M':0, 'F':1, 'T': 2}
    }
    matrix = sku.extract_sklearn_features_categorical(categories,
        categories_to_val_map, dataset)
    expected = np.asarray([
        [1, 0, 0, 0, 1, 0, 0],
        [1, 0, 0, 0, 0, 1, 0],
        [0, 1, 0, 0, 1, 0, 0],
        [0, 0, 1, 0, 1, 0, 0],
        [0, 0, 0, 1, 0, 0, 1]])
    assert np.array_equal(matrix, expected)

def test_build_categorical_to_value_map():
    dataset = pd.DataFrame({
        'Nationality':['USA', 'USA', 'France', 'Germany', 'Bengal'],
        'Gender':['M', 'F', 'M', 'M', 'T'],
        'C':['1', '2', '3', '4', '5'],
        'D':[1, None, 3, 4, 5],
        })
    columns = ['Nationality', 'Gender']
    categories_to_val_map = sku.build_categorical_to_value_map(columns, dataset)
    # Assert all the 'columns' have a codemap.
    assert set(categories_to_val_map.keys()) == set(columns)
    for col, valmap in categories_to_val_map.iteritems():
        # Assert each unique val in the column has codes for all its values.
        unique_vals = set(dataset[col].unique())
        assert unique_vals == set(valmap.keys())
        # Assert that all codes are unique.
        assert len(set(code for _, code in valmap.iteritems())) == \
            len(unique_vals)

def test_extract_sklearn_univariate_target():
    dataset = pd.DataFrame({
        'Nationality':['USA', 'USA', 'France', 'Germany', 'Bengal'],
        'Gender':['M', 'F', 'M', 'M', 'T'],
        'C':['1', '2', '3', '4', '5'],
        'D':[1, None, 3, 4, 5],
        })
    target = 'C'
    vector = sku.extract_sklearn_univariate_target(target, dataset)
    expected = ['1', '2', '3', '4', '5']
    assert np.array_equal(vector, expected)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import pytest

import bayeslite
from bayeslite.exception import BayesLiteException as BLE
from bayeslite.sqlite3_util import sqlite3_quote_name as quote

import bdbcontrib
from bdbcontrib.bql_utils import describe_generator, describe_generator_models
from bdbcontrib.metamodels.composer import Composer
from bdbcontrib.predictors import random_forest
from bdbcontrib.predictors import keplers_law
from bdbcontrib.predictors import multiple_regression


# Use satellites for all tests.
PATH_TESTS = os.path.dirname(os.path.abspath(__file__))
PATH_ROOT = os.path.dirname(PATH_TESTS)
PATH_EXAMPLES = os.path.join(PATH_ROOT, 'examples')
PATH_SATELLITES = os.path.join(PATH_EXAMPLES, 'satellites')
PATH_SATELLITES_CSV = os.path.join(PATH_SATELLITES, 'satellites.csv')

# ------------------------------------------------------------------------------
# The following live outside the TestComposer class since we do not need to
# reuse a bdb instance across tests.

def test_create_generator_schema():
    bdb = bayeslite.bayesdb_open()
    bayeslite.bayesdb_read_csv_file(bdb, 'satellites', PATH_SATELLITES_CSV,
        header=True, create=True)
    composer = Composer(n_samples=5)
    bayeslite.bayesdb_register_metamodel(bdb, composer)
    # Using crosscat and default to specify models should work.
    bdb.execute('''
        CREATE GENERATOR t1 FOR satellites USING composer(
            default (
                Country_of_Operator CATEGORICAL, Operator_Owner CATEGORICAL,
                Users CATEGORICAL, Purpose CATEGORICAL,
                Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
                Apogee_km NUMERICAL, Eccentricity NUMERICAL
            ),
            crosscat (
                Anticipated_Lifetime NUMERICAL, Contractor CATEGORICAL
            )
        );''')
    assert bayeslite.core.bayesdb_has_generator(bdb, 't1_cc')
    # IGNORE and GUESS(*) are forbidden and should crash.
    with pytest.raises(AttributeError):
        bdb.execute('''
            CREATE GENERATOR t2 FOR satellites USING composer(
                default (
                    GUESS(*), Country_of_Operator IGNORE,
                    Apogee_km NUMERICAL, Eccentricity NUMERICAL
                )
            );''')
    # Test unregistered foreign predictor.
    with pytest.raises(BLE):
        bdb.execute('''
            CREATE GENERATOR t3 FOR satellites USING composer(
                default (
                    Country_of_Operator CATEGORICAL, Operator_Owner CATEGORICAL,
                    Users CATEGORICAL, Purpose CATEGORICAL,
                    Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
                    Apogee_km NUMERICAL, Eccentricity NUMERICAL
                ),
                random_forest (
                    Apogee_km NUMERICAL GIVEN Operator_Owner
                )
            );''')
    # Unregistered foreign predictor should crash.
    with pytest.raises(BLE):
        bdb.execute('''
            CREATE GENERATOR t4 FOR satellites USING composer(
                default (
                    Country_of_Operator CATEGORICAL, Operator_Owner CATEGORICAL,
                    Users CATEGORICAL, Purpose CATEGORICAL,
                    Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
                    Apogee_km NUMERICAL, Eccentricity NUMERICAL
                ),
                random_forest (
                    Apogee_km NUMERICAL GIVEN Operator_Owner
                )
            );''')
    # Registered foreign predictor should work.
    composer.register_foreign_predictor(random_forest.RandomForest)
    bdb.execute('''
        CREATE GENERATOR t5 FOR satellites USING composer(
            default (
                Country_of_Operator CATEGORICAL, Operator_Owner CATEGORICAL,
                Users CATEGORICAL, Purpose CATEGORICAL,
                Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
                Eccentricity NUMERICAL
            ),
            random_forest (
                Apogee_km NUMERICAL GIVEN Operator_Owner
            )
        );''')
    # Wrong stattype in predictor should crash.
    with pytest.raises(BLE):
        bdb.execute('''
            CREATE GENERATOR t6 FOR satellites USING composer(
                default (
                    Country_of_Operator CATEGORICAL, Operator_Owner CATEGORICAL,
                    Users CATEGORICAL, Purpose CATEGORICAL,
                    Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
                    Apogee_km NUMERICAL, Eccentricity NUMERICAL
                ),
                random_forest (
                    Apogee_km RADIAL GIVEN Operator_Owner
                )
            );''')
    # Missing GIVEN keyword should crash.
    with pytest.raises(BLE):
        bdb.execute('''
            CREATE GENERATOR t6 FOR satellites USING composer(
                default (
                    Country_of_Operator CATEGORICAL, Operator_Owner CATEGORICAL,
                    Users CATEGORICAL, Purpose CATEGORICAL,
                    Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
                    Apogee_km NUMERICAL, Eccentricity NUMERICAL
                ),
                random_forest (
                    Apogee_km NUMERICAL, Operator_Owner
                )
            );''')
    # Missing conditions in random forest conditions should crash.
    with pytest.raises(BLE):
        bdb.execute('''
            CREATE GENERATOR t7 FOR satellites USING composer(
                default (
                    Country_of_Operator CATEGORICAL,
                    Users CATEGORICAL, Purpose CATEGORICAL,
                    Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
                    Apogee_km NUMERICAL, Eccentricity NUMERICAL
                ),
                random_forest (
                    Apogee_km NUMERICAL GIVEN Operator_Owner
                )
            );''')
    # Test duplicate declarations.
    with pytest.raises(BLE):
        bdb.execute('''
            CREATE GENERATOR t7 FOR satellites USING composer(
                default (
                    Country_of_Operator CATEGORICAL,
                    Users CATEGORICAL, Purpose CATEGORICAL,
                    Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
                    Apogee_km NUMERICAL, Eccentricity NUMERICAL
                ),
                random_forest (
                    Class_of_orbit CATEGORICAL GIVEN Operator_Owner
                )
            );''')
    # Arbitrary DAG with foreign predictors.
    composer.register_foreign_predictor(multiple_regression.MultipleRegression)
    bdb.execute('''
        CREATE GENERATOR t8 FOR satellites USING composer(
            default (
                Country_of_Operator CATEGORICAL, Operator_Owner CATEGORICAL,
                Users CATEGORICAL, Purpose CATEGORICAL,
                Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
            ),
            random_forest (
                Apogee_km NUMERICAL GIVEN Operator_Owner, Users
            ),
            multiple_regression (
                Eccentricity NUMERICAL GIVEN Apogee_km, Users, Perigee_km
            )
        );''')
    # Duplicate declarations in foreign predictors should crash.
    with pytest.raises(BLE):
        bdb.execute('''
            CREATE GENERATOR t9 FOR satellites USING composer(
                default (
                    Country_of_Operator CATEGORICAL, Operator_Owner CATEGORICAL,
                    Users CATEGORICAL, Purpose CATEGORICAL,
                    Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
                    Apogee_km NUMERICAL, Eccentricity NUMERICAL
                ),
                random_forest (
                    Perigee_km NUMERICAL GIVEN Purpose
                ),
                multiple_regression (
                    Perigee_km NUMERICAL GIVEN Operator_Owner
                )
            );''')
    # MML for default models should work.
    bdb.execute('''
        CREATE GENERATOR t10 FOR satellites USING composer(
            default (
                Country_of_Operator CATEGORICAL,
                Users CATEGORICAL, Purpose CATEGORICAL,
                Class_of_orbit CATEGORICAL, Apogee_km NUMERICAL
            )
            random_forest (
                Perigee_km NUMERICAL GIVEN Purpose
            )
            multiple_regression (
                Eccentricity NUMERICAL GIVEN Operator_Owner, Class_of_orbit
            )
            DEPENDENT(Apogee_km, Perigee_km, Purpose),
            INDEPENDENT(Country_of_Operator, Purpose)
        );''')
    # MML for foreign predictors should crash.
    with pytest.raises(BLE):
        bdb.execute('''
            CREATE GENERATOR t11 FOR satellites USING composer(
                default (
                    Country_of_Operator CATEGORICAL, Operator_Owner CATEGORICAL,
                    Users CATEGORICAL, Purpose CATEGORICAL,
                    Class_of_orbit CATEGORICAL,
                    Apogee_km NUMERICAL
                ),
                random_forest (
                    Perigee_km NUMERICAL GIVEN Purpose
                ),
                multiple_regression (
                    Eccentricity NUMERICAL GIVEN Operator_Owner, Class_of_orbit
                )
                DEPENDENT(Apogee_km, Eccentricity, Country_of_Operator),
                INDEPENDENT(Perigee_km, Purpose)
            );''')
    # Test full generator.
    composer.register_foreign_predictor(keplers_law.KeplersLaw)
    bdb.execute('''
        CREATE GENERATOR t12 FOR satellites USING composer(
            default (
                Country_of_Operator CATEGORICAL, Operator_Owner CATEGORICAL,
                Users CATEGORICAL, Purpose CATEGORICAL,
                Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
                Apogee_km NUMERICAL, Eccentricity NUMERICAL,
                Launch_Mass_kg NUMERICAL, Dry_Mass_kg NUMERICAL,
                Power_watts NUMERICAL, Date_of_Launch NUMERICAL,
                Contractor CATEGORICAL,
                Country_of_Contractor CATEGORICAL, Launch_Site CATEGORICAL,
                Launch_Vehicle CATEGORICAL,
                Source_Used_for_Orbital_Data CATEGORICAL,
                longitude_radians_of_geo NUMERICAL,
                Inclination_radians NUMERICAL,
            ),
            random_forest (
                Type_of_Orbit CATEGORICAL
                    GIVEN Apogee_km, Perigee_km,
                        Eccentricity, Period_minutes, Launch_Mass_kg,
                        Power_watts, Anticipated_Lifetime, Class_of_orbit
            ),
            keplers_law (
                Period_minutes NUMERICAL
                    GIVEN Perigee_km, Apogee_km
            ),
            multiple_regression (
                Anticipated_Lifetime NUMERICAL
                    GIVEN Dry_Mass_kg, Power_watts, Launch_Mass_kg, Contractor
            ),
            DEPENDENT(Apogee_km, Perigee_km, Eccentricity),
            INDEPENDENT(Country_of_Operator, longitude_radians_of_geo)
        );''')
    bdb.close()

def test_register():
    bdb = bayeslite.bayesdb_open()
    composer = Composer(n_samples=5)
    bayeslite.bayesdb_register_metamodel(bdb, composer)
    # Check if globally registered.
    try:
        bdb.sql_execute('''
            SELECT * FROM bayesdb_metamodel WHERE name={}
        '''.format(quote(composer.name()))).next()
    except StopIteration:
        pytest.fail('Composer not registered in bayesdb_metamodel.')
    # Check all tables/triggers.
    schema = [
        ('table', 'bayesdb_composer_cc_id'),
        ('table', 'bayesdb_composer_column_owner'),
        ('table', 'bayesdb_composer_column_toposort'),
        ('trigger', 'bayesdb_composer_column_toposort_check'),
        ('table', 'bayesdb_composer_column_parents'),
        ('table', 'bayesdb_composer_column_foreign_predictor'),
        ('trigger', 'bayesdb_composer_column_foreign_predictor_check')
    ]
    for kind, name in schema:
        try:
            bdb.sql_execute('''
                SELECT * FROM sqlite_master WHERE type={} AND name={}
            '''.format(quote(kind), quote(name))).next()
        except StopIteration:
            pytest.fail('Missing from Composer schema: {}'.format((kind,name)))
    bdb.close()

def test_register_foreign_predictor():
    bdb = bayeslite.bayesdb_open()
    composer = Composer(n_samples=5)
    bayeslite.bayesdb_register_metamodel(bdb, composer)
    # Register valid predictors.
    composer.register_foreign_predictor(random_forest.RandomForest)
    composer.register_foreign_predictor(multiple_regression.MultipleRegression)
    composer.register_foreign_predictor(keplers_law.KeplersLaw)
    # Register duplicates.
    with pytest.raises(BLE):
        composer.register_foreign_predictor(keplers_law.KeplersLaw)
    with pytest.raises(BLE):
        composer.register_foreign_predictor(
            multiple_regression.MultipleRegression)
    with pytest.raises(BLE):
        composer.register_foreign_predictor(random_forest.RandomForest)
    # Register invalid predictors.
    with pytest.raises(AssertionError):
        composer.register_foreign_predictor(None)
    with pytest.raises(AssertionError):
        composer.register_foreign_predictor('bans')


def test_drop_generator():
    bdb = bayeslite.bayesdb_open()
    # Initialize the database
    bayeslite.bayesdb_read_csv_file(bdb, 'satellites', PATH_SATELLITES_CSV,
        header=True, create=True)
    composer = Composer(n_samples=5)
    bayeslite.bayesdb_register_metamodel(bdb, composer)
    composer.register_foreign_predictor(random_forest.RandomForest)
    composer.register_foreign_predictor(multiple_regression.MultipleRegression)
    composer.register_foreign_predictor(keplers_law.KeplersLaw)
    bdb.execute('''
        CREATE GENERATOR t1 FOR satellites USING composer(
            default (
                Country_of_Operator CATEGORICAL, Operator_Owner CATEGORICAL,
                Users CATEGORICAL, Purpose CATEGORICAL,
                Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
                Apogee_km NUMERICAL, Eccentricity NUMERICAL,
                Launch_Mass_kg NUMERICAL, Dry_Mass_kg NUMERICAL,
                Power_watts NUMERICAL, Date_of_Launch NUMERICAL,
                Contractor CATEGORICAL,
                Country_of_Contractor CATEGORICAL, Launch_Site CATEGORICAL,
                Launch_Vehicle CATEGORICAL,
                Source_Used_for_Orbital_Data CATEGORICAL,
                longitude_radians_of_geo NUMERICAL,
                Inclination_radians NUMERICAL,
            ),
            random_forest (
                Type_of_Orbit CATEGORICAL
                    GIVEN Apogee_km, Perigee_km,
                        Eccentricity, Period_minutes, Launch_Mass_kg,
                        Power_watts, Anticipated_Lifetime, Class_of_orbit
            ),
            keplers_law (
                Period_minutes NUMERICAL
                    GIVEN Perigee_km, Apogee_km
            ),
            multiple_regression (
                Anticipated_Lifetime NUMERICAL
                    GIVEN Dry_Mass_kg, Power_watts, Launch_Mass_kg,
                    Contractor
            ),
            DEPENDENT(Apogee_km, Perigee_km, Eccentricity),
            DEPENDENT(Contractor, Country_of_Contractor),
            INDEPENDENT(Country_of_Operator, Date_of_Launch)
        );''')
    generator_id = bayeslite.core.bayesdb_get_generator(bdb, 't1')
    schema = [
        ('table', 'bayesdb_composer_cc_id'),
        ('table', 'bayesdb_composer_column_owner'),
        ('table', 'bayesdb_composer_column_toposort'),
        ('table', 'bayesdb_composer_column_parents'),
        ('table', 'bayesdb_composer_column_foreign_predictor'),
    ]
    # Iterate through tables before dropping.
    for _, name in schema:
        bdb.sql_execute('''
            SELECT * FROM {} WHERE generator_id=?
        '''.format(quote(name)), (generator_id,)).next()
    # Drop generator and ensure table lookups with generator_id throw error.
    bdb.execute('DROP GENERATOR t1')
    for _, name in schema:
        with pytest.raises(StopIteration):
            bdb.sql_execute('''
                SELECT * FROM {} WHERE generator_id=?
            '''.format(quote(name)), (generator_id,)).next()
    assert not bayeslite.core.bayesdb_has_generator(bdb, 't1')
    assert not bayeslite.core.bayesdb_has_generator(bdb, 't1_cc')
    bdb.close()

def test_composer_integration__ci_slow():
    # But currently difficult to seperate these tests into smaller tests because
    # of their sequential nature. We will still test all internal functions
    # with different regimes of operation.

    # SETUP
    # -----
    # Dataset.
    bdb = bayeslite.bayesdb_open()
    bayeslite.bayesdb_read_csv_file(bdb, 'satellites', PATH_SATELLITES_CSV,
        header=True, create=True)
    bdbcontrib.bql_utils.nullify(bdb, 'satellites', 'NaN')
    # Composer.
    composer = Composer(n_samples=5)
    composer.register_foreign_predictor(
        multiple_regression.MultipleRegression)
    composer.register_foreign_predictor(keplers_law.KeplersLaw)
    composer.register_foreign_predictor(random_forest.RandomForest)
    # Use complex generator for interesting test cases.
    bayeslite.bayesdb_register_metamodel(bdb, composer)
    bdb.execute('''
        CREATE GENERATOR t1 FOR satellites USING composer(
            default (
                Country_of_Operator CATEGORICAL, Operator_Owner CATEGORICAL,
                Users CATEGORICAL, Purpose CATEGORICAL,
                Class_of_orbit CATEGORICAL, Perigee_km NUMERICAL,
                Apogee_km NUMERICAL, Eccentricity NUMERICAL,
                Launch_Mass_kg NUMERICAL, Dry_Mass_kg NUMERICAL,
                Power_watts NUMERICAL, Date_of_Launch NUMERICAL,
                Contractor CATEGORICAL,
                Country_of_Contractor CATEGORICAL, Launch_Site CATEGORICAL,
                Launch_Vehicle CATEGORICAL,
                Source_Used_for_Orbital_Data CATEGORICAL,
                longitude_radians_of_geo NUMERICAL,
                Inclination_radians NUMERICAL,
            ),
            random_forest (
                Type_of_Orbit CATEGORICAL
                    GIVEN Apogee_km, Perigee_km,
                        Eccentricity, Period_minutes, Launch_Mass_kg,
                        Power_watts, Anticipated_Lifetime, Class_of_orbit
            ),
            keplers_law (
                Period_minutes NUMERICAL
                    GIVEN Perigee_km, Apogee_km
            ),
            multiple_regression (
                Anticipated_Lifetime NUMERICAL
                    GIVEN Dry_Mass_kg, Power_watts, Launch_Mass_kg,
                    Contractor
            ),
            DEPENDENT(Apogee_km, Perigee_km, Eccentricity),
            DEPENDENT(Contractor, Country_of_Contractor),
            INDEPENDENT(Country_of_Operator, Date_of_Launch)
        );''')


    # ----------------------
    # TEST INITIALIZE MODELS
    # ----------------------

    bdb.execute('INITIALIZE 2 MODELS FOR t1')
    # Check number of models.
    df = describe_generator_models(bdb, 't1')
    assert len(df) == 2
    df = describe_generator_models(bdb, 't1_cc')
    assert len(df) == 2

    # -------------------
    # TEST ANALYZE MODELS
    # -------------------

    bdb.execute('ANALYZE t1 FOR 2 ITERATIONS WAIT;')
    # Check number of iterations of composer.
    df = describe_generator_models(bdb, 't1')
    for index, modelno, iterations in df.itertuples():
        assert iterations == 2
    # Check number of iterations of composer_cc.
    df = describe_generator_models(bdb, 't1_cc')
    for index, modelno, iterations in df.itertuples():
        assert iterations == 2

    # ----------------------------------
    # TEST COLUMN DEPENDENCE PROBABILITY
    # ----------------------------------

    # Special 0/1 regimes.
    # Local with a INDEPENDENT local should be 0.
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Date_of_Launch
            WITH Country_of_Operator FROM t1 LIMIT 1
    ''')
    assert curs.next()[0] == 0
    # Local with a DEPENDENT local should be 1.
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Perigee_km WITH Eccentricity
            FROM t1 LIMIT 1
    ''')
    assert curs.next()[0] == 1
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Apogee_km WITH Eccentricity
            FROM t1 LIMIT 1
    ''')
    assert curs.next()[0] == 1
    # Foreign with a local parent should be 1.
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Period_minutes WITH Apogee_km
            FROM t1 LIMIT 1
    ''')
    assert curs.next()[0] == 1.
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Anticipated_Lifetime WITH Power_watts
            FROM t1 LIMIT 1
    ''')
    assert curs.next()[0] == 1.
    # Foreign with a foreign parent should be 1.
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Type_of_Orbit WITH
            Anticipated_Lifetime FROM t1 LIMIT 1
    ''')
    assert curs.next()[0] == 1.
    # Foreign with a local non-parent DEPENDENT with local parent should be 1.
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Period_minutes WITH
            Eccentricity FROM t1 LIMIT 1
    ''')
    assert curs.next()[0] == 1.
    # Foreign with foreign sharing common direct ancestor should be 1.
    # Launch_Mass_kg is the common parent.
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Anticipated_Lifetime WITH
            Type_of_Orbit FROM t1 LIMIT 1
    ''')
    assert curs.next()[0] == 1.
    # Foreign with a foreign sharing a common DEPENDENT ancestor should be 1.
    # Eccentricity is a parent of Type_of_orbit, and is dependent
    # with Period_minutes through DEPENDENT(Apogee_km, Perigee_km, Eccentricity)
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Period_minutes WITH
            Type_of_Orbit FROM t1 LIMIT 1
    ''')
    assert curs.next()[0] == 1.
    # Column with itself should be 1.
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Anticipated_Lifetime WITH
            Anticipated_Lifetime FROM t1 LIMIT 1
    ''')
    assert curs.next()[0] == 1.

    # Unknown [0,1] regimes.
    # Foreign with a local of unknown relation with parents.
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Anticipated_Lifetime WITH
            longitude_radians_of_geo FROM t1 LIMIT 1
    ''')
    assert 0 <= curs.next()[0] <= 1.
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Period_minutes WITH
            longitude_radians_of_geo FROM t1 LIMIT 1
    ''')
    assert 0 <= curs.next()[0] <= 1.
    # Foreign with a foreign of unknown ancestry relation.
    curs = bdb.execute('''
        ESTIMATE DEPENDENCE PROBABILITY OF Anticipated_Lifetime WITH
            Period_minutes FROM t1 LIMIT 1
    ''')
    assert 0 <= curs.next()[0] <= 1.

    # ----------------------------------
    # TEST SIMULATE
    # ----------------------------------

    # Crash tests for various code paths. Quality of simulations ignored.
    # Joint local.
    curs = bdb.execute('''
        SIMULATE Power_watts, Launch_Mass_kg FROM t1 LIMIT 2;
    ''')
    assert len(curs.fetchall()) == 2
    # Forward simulate foreign.
    curs = bdb.execute('''
        SIMULATE Period_minutes FROM t1 GIVEN Apogee_km = 1000, Perigee_km = 980
            LIMIT 2;
    ''')
    assert len(curs.fetchall()) == 2
    # Forward simulate foreign with missing parents.
    curs = bdb.execute('''
        SIMULATE Anticipated_Lifetime FROM t1 GIVEN Dry_Mass_kg = 2894,
            Launch_Mass_kg = 1730 LIMIT 2;
    ''')
    assert len(curs.fetchall()) == 2
    # Joint simulate foreign with parents, and missing parents.
    curs = bdb.execute('''
        SIMULATE Type_of_Orbit, Eccentricity FROM t1 GIVEN Dry_Mass_kg = 2894,
            Launch_Mass_kg = 1730 LIMIT 2;
    ''')
    assert len(curs.fetchall()) == 2
    # Joint simulate foreign with non-parents.
    curs = bdb.execute('''
        SIMULATE Period_minutes, Eccentricity FROM t1 GIVEN Apogee_km = 38000
            LIMIT 2;
    ''')
    assert len(curs.fetchall()) == 2
    # Simulate joint local conditioned on two foreigns.
    curs = bdb.execute('''
        SIMULATE Country_of_Operator, Inclination_radians FROM t1
            GIVEN Period_minutes = 1432, Anticipated_Lifetime = 5 LIMIT 2;
    ''')
    assert len(curs.fetchall()) == 2
    # Simulate joint foreign conditioned on third foreign.
    curs = bdb.execute('''
        SIMULATE Period_minutes, Anticipated_Lifetime FROM t1
            GIVEN Type_of_Orbit = 'Deep Highly Eccentric' LIMIT 2
    ''')
    assert len(curs.fetchall()) == 2
    # Simulate foreign conditioned on itself.
    curs = bdb.execute('''
        SIMULATE Period_minutes, Apogee_km FROM t1
            GIVEN Period_minutes = 102 LIMIT 2
    ''')
    assert [s[0] for s in curs] == [102] * 2

    # -----------------------------
    # TEST COLUMN VALUE PROBABILITY
    # -----------------------------

    # Crash tests for various code path. Quality of logpdf ignored.
    # Conditional local.
    curs = bdb.execute('''
        ESTIMATE PROBABILITY OF Power_watts = 800 GIVEN (Perigee_km = 980,
            Launch_Mass_kg = 890) FROM t1 LIMIT 1;
    ''')
    assert 0. <= curs.next()[0]
    # Unconditional foreign
    curs = bdb.execute('''
        ESTIMATE PROBABILITY OF Period_minutes = 1020 FROM t1 LIMIT 1;
    ''')
    assert 0. <= curs.next()[0]
    # Conditional foreign on parent and non-parents.
    curs = bdb.execute('''
        ESTIMATE PROBABILITY OF Period_minutes = 1020 GIVEN
            (Apogee_km = 38000, Eccentricity = 0.03) FROM t1 LIMIT 1;
    ''')
    assert 0 <= curs.next()[0]
    # Conditional foriegn on foreign.
    curs = bdb.execute('''
        ESTIMATE PROBABILITY OF Anticipated_Lifetime = 4.09 GIVEN
            (Class_of_Orbit = 'LEO', Purpose='Astrophysics',
                Period_minutes = 1436) FROM t1 LIMIT 1;
    ''')
    assert 0. <= curs.next()[0]
    # Categorical foreign should be less than 1.
    curs = bdb.execute('''
        ESTIMATE PROBABILITY OF Type_of_Orbit = 'Polar' FROM t1 LIMIT 1;
    ''')
    assert curs.next()[0] <= 1.
    # Query inconsistent with evidence should be 0.
    curs = bdb.execute('''
        ESTIMATE PROBABILITY OF "Type_of_Orbit" = 'Polar'
            GIVEN ("Type_of_Orbit" = 'Deep Highly Eccentric') FROM t1 LIMIT 1;
    ''')
    assert curs.next()[0] == 0.
    # In theory, query consistent with evidence should be 1, but this is very
    # hard to ensure due to stochastic sampling giving different estimates of
    # P(Y), once in joint and once in marginal Monte Carlo estimation.

    # -----------------------
    # TEST MUTUAL INFORMATION
    # -----------------------

    # Two local columns.
    curs = bdb.execute('''
        ESTIMATE MUTUAL INFORMATION OF Country_of_Contractor WITH
            longitude_radians_of_geo USING 5 SAMPLES FROM t1 LIMIT 1;
    ''')
    # XXX Small sample sizes non-deterministically produce negative MI
    assert -1 <= curs.next()[0]
    # One local and one foreign column.
    curs = bdb.execute('''
        ESTIMATE MUTUAL INFORMATION OF Period_minutes WITH
            longitude_radians_of_geo USING 5 SAMPLES FROM t1 LIMIT 1;
    ''')
    # XXX This non-deterministically fails when sample sizes are small
    # assert 0. <= curs.next()[0]
    assert float("-inf") <= curs.next()[0]
    # Two foreign columns.
    curs = bdb.execute('''
        ESTIMATE MUTUAL INFORMATION OF Period_minutes WITH
            Anticipated_Lifetime USING 5 SAMPLES FROM t1 LIMIT 1;
    ''')
    # XXX This non-deterministically fails when sample sizes are small
    # assert 0. <= curs.next()[0]
    assert float("-inf") <= curs.next()[0]

    # -----------------------
    # TEST PREDICT CONFIDENCE
    # -----------------------

    # Continuous local column.
    curs = bdb.execute('''
        INFER EXPLICIT PREDICT Dry_Mass_kg CONFIDENCE c FROM t1 LIMIT 1;
    ''')
    assert curs.next()[1] >= 0.
    # Discrete local column with no children.
    curs = bdb.execute('''
        INFER EXPLICIT PREDICT Purpose CONFIDENCE c FROM t1 LIMIT 1;
    ''')
    assert 0 <= curs.next()[1] <= 1
    # Discrete local column with children.
    curs = bdb.execute('''
        INFER EXPLICIT PREDICT Contractor CONFIDENCE c FROM t1 LIMIT 1;
    ''')
    assert 0 <= curs.next()[1] <= 1
    # Continuous foreign columns.
    curs = bdb.execute('''
        INFER EXPLICIT PREDICT Period_minutes CONFIDENCE c FROM t1 LIMIT 1;
    ''')
    assert curs.next()[1] >= 0.
    # Discrete foreign column.
    curs = bdb.execute('''
        INFER EXPLICIT PREDICT Type_of_Orbit CONFIDENCE c FROM t1 LIMIT 1;
    ''')
    assert 0 <= curs.next()[1] <= 1

    bdb.close()

def test_topological_sort():
    # Acyclic graph should return a correct topo sort.
    graph = {1:[], 2:[1], 3:[2,6], 4:[2], 5:[4], 6:[4,1]}
    topo = Composer.topological_sort(graph)
    for i in xrange(len(topo)):
        node, parents = topo[i]
        for n in topo[i+1:]:
            assert n not in parents
    # Cyclic graph should throw an error.
    graph = {1:[], 2:[1], 3:[2,4,6], 4:[2], 5:[3,4], 6:[4,5,1]}
    with pytest.raises(BLE):
        topo = Composer.topological_sort(graph)



# -*- coding: utf-8 -*-

#   Copyright (c) 2010-2016, MIT Probabilistic Computing Project
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import math

import bayeslite.core
from bayeslite import bql_quote_name
from bayeslite.exception import BayesLiteException as BLE

def extract_target_cols(bdb, generator, targets=None):
    """Extract target columns (helper for LL/KL query).

    If targets is None, then a list of all sqlite3 quoted column names from
    generator are returned.

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        Active BayesDB instance.
    generator : str
        Name of generator.
    targets : list<str>, optional
        List of columns in the table.

    Returns
    -------
    target_cols : list<str>
    """
    if targets is None:
        generator_id = bayeslite.core.bayesdb_get_generator(bdb, generator)
        targets = bayeslite.core.bayesdb_generator_column_names(bdb,
            generator_id)
    return map(bql_quote_name, targets)


def extract_given_cols_vals(givens=None):
    """Extract given columns and values (helper for LL/KL).

    If givens is None, then an empty list is returned. Otherwise an appropriate
    list of tuples is returned, where the first element is the sqlite3 quoted
    name and the second element is the constraint value.

    Parameters
    ----------
    givens : list<str>, optional
        List of columns in the table.

    Returns
    -------
    given_cols_vals : list<tuple>
    """
    if givens is None:
        return []

    given_cols = map(bql_quote_name, givens[::2])
    given_vals = givens[1::2]
    assert len(given_cols) == len(given_vals)
    return zip(given_cols, given_vals)


def estimate_log_likelihood(bdb, table, generator, targets=None, givens=None,
        n_samples=None):
    """Estimate the log likelihood for obsevations in a table.

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        Active BayesDB instance.
    table : str
        Name of table.
    generator : str
        Name of generator.
    targets : list<str>, optional
        List of columns in the table for which to compute the log-likelihood.
        Defaults to all the columns.
    givens : list<tuple>, optional
        A list of [(column, value)] pairs on which to condition on. Defaults to
        no conditionals. See example for more details.
    n_samples : int, optional
        Number of rows from table to use in the computation. Defaults to all
        the rows.

    Returns
    -------
    ll : float
        The log likelihood of the table[columns] under the conditional
        distribution (specified by givens) of generator.

    Example:
    estimate_log_likelihood(bdb, 'people', 'people_gen',
        targets=['weight', 'height'],
        givens=[('nationality', 'USA'), ('age', 17)])
    """
    # Defaults to all columns if targets is None.
    targets = extract_target_cols(bdb, generator, targets=targets)

    # Defaults to no givens if givens is None
    givens = extract_given_cols_vals(givens=givens)
    givens = ','.join(['{}={}'.format(c,v) for (c,v) in givens])

    # Obtain the dataset table.
    table = bql_quote_name(table.strip(';'))
    sql = '''
        SELECT {} FROM {};
    '''.format(','.join(targets), table)
    dataset = bdb.execute(sql)

    # Obtain number of rows in the dataset and samples to use.
    n_samples = n_samples
    n_rows = bdb.execute('''
        SELECT COUNT(*) FROM {}'''.format(table)).fetchvalue()
    if n_samples is None or n_rows < n_samples:
        n_samples = n_rows

    # Compute the log-likelihood of the targets, subject to givens.
    # XXX This code is currently wrong due to shortcomings in BQL:
    #  - BQL cannot evaluate joint density. Assume that all the rows are IID,
    #  and that all the columns factor into their marginal density.
    ll, i = 0, 0
    for row in dataset:
        if i > n_samples:
            break
        else:
            i += 1
        # XXX Wrong: assume joint factors into product of marginals.
        for col, val in zip(targets, row):
            if givens:
                # XXX TODO write GIVEN in this query using bindings.
                bql = '''
                    ESTIMATE PROBABILITY OF {}=? GIVEN ({}) FROM {} LIMIT 1
                '''.format(col, givens, bql_quote_name(generator))
            else:
                bql = '''
                    ESTIMATE PROBABILITY OF {}=? FROM {} LIMIT 1
                '''.format(col, bql_quote_name(generator))

            ll += math.log(bdb.execute(bql, (val,)).fetchvalue())

    return ll

def estimate_kl_divergence(bdb, generatorA, generatorB, targets=None,
        givens=None, n_samples=None):
    """Estimate the KL divergence.

    The KL divergence is a mesaure of the "information lost" when generatorB
    (the approximating generator) is used to approximate generatorA (the base
    generator). KL divergence is not symmetric in, and KL(genA||genB) is not
    necessarily equal to KL(genB||genA).

    TODO: Monte Carlo estimation is a terrible way to compute the KL divergence.
    (Not to say there are better methods in general). One illustration of this
    is that the estimated KL divergence has emperically been shown to obtain
    negative realizations for high-dimensional data.

    Computing the KL divergence in general (of high dimensional distributions)
    is a very hard problem; most research uses the structure of the
    distributions to find good estimators. Adaptive quadrature or exact methods
    for numerical integration could outperform Monte Carlo?

    TODO: More sophisticated algorithm for detecting cases where absolute
    continuity could be a problem (currently have a heuristic).
    As it stands, Monte Carlo estimates may have infinite variance depending
    on simulated values from generatorA.

    Parameters
    ----------
    bdb : bayeslite.BayesDB
        Active BayesDB instance.
    generatorA : str
        Name of base generator.
    generatorB : str
        Name of approximating generator.
    targets : list<str>, optional
        List of columns in the table for which to compute the log-likelihood.
        Defaults to all the columns.
    givens : list<tuple>, optional
        A list of [(column, value)] pairs on which to condition on. Defaults to
        no conditionals. See example for more details.
    n_samples: int, optional
        Number of simulated samples to use in the Monte Carlo estimate.

    Returns
    -------
    kl : float
        The KL divergence. May be infinity.

    Example:
    estimate_kl_divergence(bdb, 'crosscat_gen', 'baxcat_gen',
        targets=['weight', 'height'],
        givens=[('nationality', 'USA'), ('age', 17)])
    """
    # XXX Default to 10,000 samples
    if n_samples is None:
        n_samples = 10000

    # Defaults to all columns if targets is None.
    targets = extract_target_cols(bdb, generatorA, targets=targets)

    # Defaults to no givens if givens is None
    givens = extract_given_cols_vals(givens=givens)
    givens = ','.join(['{}={}'.format(c,v) for (c,v) in givens])

    # Obtain samples from the base distribution.
    if givens:
        # XXX TODO write GIVEN in this query using bindings.
        bql = '''
            SIMULATE {} FROM {} GIVEN {} LIMIT {}
        '''.format(','.join(targets), bql_quote_name(generatorA),
            givens, n_samples)
    else:
        bql = '''
            SIMULATE {} FROM {} LIMIT {}
        '''.format(','.join(targets), bql_quote_name(generatorA),
                    n_samples)
    samples = bdb.execute(bql)

    kl = 0
    for s in samples:
        logp_a, logp_b = 0, 0
        # XXX Assume joint probability factors by summing univariate
        # (conditional) probability of each cell value. This is clearly wrong,
        # until we can evaluate joint densities in BQL.
        for col, val in zip(targets, s):
            bql = '''
                ESTIMATE PROBABILITY OF {}=? FROM {} LIMIT 1
            '''.format(col, bql_quote_name(generatorA))
            crs = bdb.execute(bql, (val,))
            p_a = crs.fetchvalue()

            bql = '''
                ESTIMATE PROBABILITY OF {}=? FROM {} LIMIT 1
            '''.format(col, bql_quote_name(generatorB))
            crs = bdb.execute(bql, (val,))
            p_b = crs.fetchvalue()

            # XXX Heuristic to detect when genA is not absolutely
            # continuous wrt genB
            if p_a == 0:
                # How on earth did we simulate a value from genA with zero
                # density/prob under genA?
                raise BLE(ValueError(
                    'Fatal error: simulated a (col,val)=({},{}) '
                    'from base generatorA ({}) with zero density. Check '
                    'implementation of simluate and/or logpdf of '
                    'generator.'.format(col,val,generatorA)))
            if p_b == 0:
                # Detected failure of absolute continuity
                # (under assumption that joint factors into marginals)
                return float('inf')

            logp_a += math.log(p_a)
            logp_b += math.log(p_b)

        kl += (logp_a - logp_b)

    # XXX Assertion may fail, see TODO in docstring.
    # assert kl > 0
    if kl < 0:
        raise BLE(ValueError(
            'Cannot compute reasonable value for KL divergence. '
            'Try increasing the number of samples (currently using {}'
            'samples).'.format(n_samples)))

    return kl / n_samples


# TODO: Migrate from hooks/contrib_diagnostics. Need users run experiments?
# def run_bdb_experiment(bdb, exp_args):
    # pass
